@@ -1,47 +1,51 @@
 from __future__ import annotations
 
-from typing import Generic, Sequence, TypeVar
+from collections.abc import Sequence
+from enum import auto
 
 from xdsl.dialects import memref
 from xdsl.dialects.builtin import (
-    ArrayAttr,
+    DYNAMIC_INDEX,
+    AffineMapAttr,
     DenseArrayBase,
     FunctionType,
     IndexType,
-    IntegerAttr,
     StringAttr,
+    SymbolNameConstraint,
     SymbolRefAttr,
     UnitAttr,
     i32,
+    i64,
 )
 from xdsl.ir import (
     Attribute,
+    Block,
     Dialect,
+    EnumAttribute,
     Operation,
-    OpResult,
     ParametrizedAttribute,
     Region,
+    SpacedOpaqueSyntaxAttribute,
     SSAValue,
+    StrEnum,
     TypeAttribute,
 )
-from xdsl.ir.core import Block
 from xdsl.irdl import (
+    AnyOf,
     AttrSizedOperandSegments,
     IRDLOperation,
-    Operand,
-    OptOperand,
-    OptOpResult,
-    ParameterDef,
-    VarOperand,
     attr_def,
     irdl_attr_definition,
     irdl_op_definition,
     operand_def,
     opt_attr_def,
     opt_operand_def,
+    opt_prop_def,
     opt_result_def,
+    prop_def,
     region_def,
     result_def,
+    traits_def,
     var_operand_def,
 )
 from xdsl.parser import AttrParser
@@ -50,7 +54,7 @@ from xdsl.traits import (
     HasParent,
     IsolatedFromAbove,
     IsTerminator,
-    SingleBlockImplicitTerminator,
+    NoTerminator,
     SymbolOpInterface,
     SymbolTable,
 )
@@ -62,107 +66,99 @@ class AsyncTokenType(ParametrizedAttribute, TypeAttribute):
     name = "gpu.async.token"
 
 
-@irdl_attr_definition
-class _AllReduceOperationAttr(ParametrizedAttribute):
-    name = "all_reduce_op"
+class AllReduceOpEnum(StrEnum):
+    Add = auto()
+    And = auto()
+    Max = auto()
+    Min = auto()
+    Mul = auto()
+    Or = auto()
+    Xor = auto()
 
-    param: ParameterDef[StringAttr]
+
+class DimensionEnum(StrEnum):
+    X = auto()
+    Y = auto()
+    Z = auto()
+
+
+class ProcessorEnum(StrEnum):
+    Sequential = auto()
+    Block_X = auto()
+    Block_Y = auto()
+    Block_Z = auto()
+    Thread_X = auto()
+    Thread_Y = auto()
+    Thread_Z = auto()
+
+
+@irdl_attr_definition
+class AllReduceOpAttr(EnumAttribute[AllReduceOpEnum], SpacedOpaqueSyntaxAttribute):
+    name = "gpu.all_reduce_op"
+
+
+@irdl_attr_definition
+class DimensionAttr(EnumAttribute[DimensionEnum], SpacedOpaqueSyntaxAttribute):
+    name = "gpu.dim"
+
+
+@irdl_attr_definition
+class ProcessorAttr(EnumAttribute[ProcessorEnum], SpacedOpaqueSyntaxAttribute):
+    name = "gpu.processor"
+
+
+@irdl_attr_definition
+class LoopDimMapAttr(ParametrizedAttribute):
+    name = "gpu.loop_dim_map"
+
+    processor: ProcessorAttr
+    map: AffineMapAttr
+    bound: AffineMapAttr
 
     def print_parameters(self, printer: Printer) -> None:
-        printer.print(f"all_reduce_op {self.param.data}")
-
-
-@irdl_attr_definition
-class _DimensionAttr(ParametrizedAttribute):
-    name = "dim"
-
-    param: ParameterDef[StringAttr]
-
-    def print_parameters(self, printer: Printer) -> None:
-        printer.print(f"dim {self.param.data}")
-
-
-T = TypeVar("T", bound=_AllReduceOperationAttr | _DimensionAttr, covariant=True)
-
-
-@irdl_attr_definition
-class _GPUAttr(ParametrizedAttribute, Generic[T]):
-    name = "gpu"
-
-    value: ParameterDef[T]
+        with printer.in_angle_brackets():
+            printer.print_string("processor = ")
+            printer.print_string(self.processor.data)
+            printer.print_string(", map = ")
+            printer.print_string(str(self.map.data))
+            printer.print_string(", bound = ")
+            printer.print_string(str(self.bound.data))
 
     @classmethod
-    def parse_parameters(cls, parser: AttrParser) -> list[Attribute]:
-        parser.parse_characters(
-            "<",
-            ": gpu attributes currently have the #gpu<name value> syntax.",
-        )
-        if parser.parse_optional_keyword("dim"):
-            attrtype = _DimensionAttr
-            vtok = parser.parse_optional_identifier()
-            if vtok not in ["x", "y", "z"]:
-                parser.raise_error(
-                    f"Unexpected dim {vtok}. A gpu dim can only be x, y, or z",
-                )
-
-        elif parser.parse_optional_keyword("all_reduce_op"):
-            attrtype = _AllReduceOperationAttr
-            vtok = parser.parse_optional_identifier()
-            if vtok not in ["add", "and", "max", "min", "mul", "or", "xor"]:
-                parser.raise_error(
-                    f"Unexpected op {vtok}. A gpu all_reduce_op can only be add, "
-                    "and, max, min, mul, or, or xor ",
-                )
-        else:
-            parser.raise_error("'dim' or 'all_reduce_op' expected")
-        parser.parse_characters(
-            ">",
-            ". gpu attributes currently have the #gpu<name value> syntax.",
-        )
-        return [attrtype([StringAttr(vtok)])]
-
-    @staticmethod
-    def from_op(value: str) -> AllReduceOperationAttr:
-        return AllReduceOperationAttr([_AllReduceOperationAttr([StringAttr(value)])])
-
-    @property
-    def data(self) -> str:
-        return self.value.param.data
-
-    @staticmethod
-    def from_dimension(value: str) -> DimensionAttr:
-        return DimensionAttr([_DimensionAttr([StringAttr(value)])])
-
-    def print_parameters(self, printer: Printer) -> None:
-        printer.print_string("<")
-        self.value.print_parameters(printer)
-        printer.print_string(">")
-
-
-DimensionAttr = _GPUAttr[_DimensionAttr]
-AllReduceOperationAttr = _GPUAttr[_AllReduceOperationAttr]
-
-_Element = TypeVar("_Element", bound=Attribute, covariant=True)
+    def parse_parameters(cls, parser: AttrParser):
+        with parser.in_angle_brackets():
+            parser.parse_keyword("processor")
+            parser.parse_punctuation("=")
+            proc = ProcessorAttr.parse_parameter(parser)
+            processor = ProcessorAttr(proc)
+            parser.parse_punctuation(",")
+            parser.parse_keyword("map")
+            parser.parse_punctuation("=")
+            map = AffineMapAttr(parser.parse_affine_map())
+            parser.parse_punctuation(",")
+            parser.parse_keyword("bound")
+            parser.parse_punctuation("=")
+            bound = AffineMapAttr(parser.parse_affine_map())
+        return [processor, map, bound]
 
 
 @irdl_op_definition
 class AllocOp(IRDLOperation):
     name = "gpu.alloc"
-    hostShared: UnitAttr | None = opt_attr_def(UnitAttr)
-    asyncDependencies: VarOperand = var_operand_def(AsyncTokenType)
-    dynamicSizes: VarOperand = var_operand_def(IndexType)
-    symbolOperands: VarOperand = var_operand_def(IndexType)
+    hostShared = opt_attr_def(UnitAttr)
+    asyncDependencies = var_operand_def(AsyncTokenType)
+    dynamicSizes = var_operand_def(IndexType)
+    symbolOperands = var_operand_def(IndexType)
 
-    irdl_options = [AttrSizedOperandSegments()]
+    irdl_options = [AttrSizedOperandSegments(as_property=True)]
 
-    result: OpResult = result_def(memref.MemRefType[Attribute])
-    asyncToken: OptOpResult = opt_result_def(AsyncTokenType)
+    result = result_def(memref.MemRefType)
+    asyncToken = opt_result_def(AsyncTokenType)
 
     def verify_(self) -> None:
         ndyn = len(self.dynamicSizes)
-        assert isinstance(self.result.type, memref.MemRefType)
-        res_type: memref.MemRefType[Attribute] = self.result.type
-        ndyn_type = len([i for i in res_type.shape.data if i.value.data == -1])
+        assert isinstance(res_type := self.result.type, memref.MemRefType)
+        ndyn_type = len([i for i in res_type.get_shape() if i == DYNAMIC_INDEX])
         if ndyn != ndyn_type:
             raise VerifyException(
                 f"Expected {ndyn_type} dynamic sizes, got {ndyn}. All "
@@ -171,7 +167,7 @@ class AllocOp(IRDLOperation):
 
     def __init__(
         self,
-        return_type: memref.MemRefType[_Element],
+        return_type: memref.MemRefType,
         dynamic_sizes: Sequence[SSAValue | Operation] | None = None,
         host_shared: bool = False,
         async_dependencies: Sequence[SSAValue | Operation] | None = None,
@@ -187,7 +183,7 @@ class AllocOp(IRDLOperation):
         attributes: dict[str, Attribute] = (
             {"hostShared": UnitAttr()} if host_shared else {}
         )
-        return super().__init__(
+        super().__init__(
             operands=[async_dependencies_vals, dynamic_sizes_vals, []],
             result_types=[return_type, token_return],
             attributes=attributes,
@@ -197,24 +193,24 @@ class AllocOp(IRDLOperation):
 @irdl_op_definition
 class AllReduceOp(IRDLOperation):
     name = "gpu.all_reduce"
-    op: AllReduceOperationAttr | None = opt_attr_def(AllReduceOperationAttr)
-    uniform: UnitAttr | None = opt_attr_def(UnitAttr)
-    operand: Operand = operand_def(Attribute)
-    result: OpResult = result_def(Attribute)
-    body: Region = region_def()
+    op = opt_prop_def(AllReduceOpAttr)
+    uniform = opt_prop_def(UnitAttr)
+    operand = operand_def(Attribute)
+    result = result_def(Attribute)
+    body = region_def()
 
-    traits = frozenset([IsolatedFromAbove()])
+    traits = traits_def(IsolatedFromAbove())
 
     @staticmethod
     def from_op(
-        op: AllReduceOperationAttr,
+        op: AllReduceOpAttr,
         operand: SSAValue | Operation,
         uniform: UnitAttr | None = None,
     ):
         return AllReduceOp.build(
             operands=[operand],
             result_types=[SSAValue.get(operand).type],
-            attributes={
+            properties={
                 "op": op,
                 "uniform": uniform,
             },
@@ -228,7 +224,7 @@ class AllReduceOp(IRDLOperation):
         return AllReduceOp.build(
             operands=[operand],
             result_types=[SSAValue.get(operand).type],
-            attributes={"uniform": uniform} if uniform is not None else {},
+            properties={"uniform": uniform} if uniform is not None else {},
             regions=[body],
         )
 
@@ -239,7 +235,7 @@ class AllReduceOp(IRDLOperation):
                 f"{self.operand.type}. They must be the same type for gpu.all_reduce"
             )
 
-        non_empty_body = any(b.ops for b in self.body.blocks)
+        non_empty_body = bool(self.body.blocks)
         op_attr = self.op is not None
         if non_empty_body == op_attr:
             if op_attr:
@@ -252,9 +248,8 @@ class AllReduceOp(IRDLOperation):
                     "gpu.all_reduce need either a non empty body or an op attribute."
                 )
         if non_empty_body:
-            region_args = self.body.blocks[0].args
-            args_types = [r.type for r in region_args]
-            if args_types != [self.result.type, self.operand.type]:
+            args_types = self.body.blocks[0].arg_types
+            if args_types != (self.result.type, self.operand.type):
                 raise VerifyException(
                     f"Expected {[str(t) for t in [self.result.type, self.operand.type]]}, "
                     f"got {[str(t) for t in args_types]}. A gpu.all_reduce's body must "
@@ -267,43 +262,39 @@ class BarrierOp(IRDLOperation):
     name = "gpu.barrier"
 
     def __init__(self):
-        return super().__init__()
+        super().__init__()
 
 
 @irdl_op_definition
 class BlockDimOp(IRDLOperation):
     name = "gpu.block_dim"
-    dimension: DimensionAttr = attr_def(DimensionAttr)
-    result: OpResult = result_def(IndexType)
+    dimension = prop_def(DimensionAttr)
+    result = result_def(IndexType)
 
     def __init__(self, dim: DimensionAttr):
-        return super().__init__(
-            result_types=[IndexType()], attributes={"dimension": dim}
-        )
+        super().__init__(result_types=[IndexType()], properties={"dimension": dim})
 
 
 @irdl_op_definition
 class BlockIdOp(IRDLOperation):
     name = "gpu.block_id"
-    dimension: DimensionAttr = attr_def(DimensionAttr)
-    result: OpResult = result_def(IndexType)
+    dimension = prop_def(DimensionAttr)
+    result = result_def(IndexType)
 
     def __init__(self, dim: DimensionAttr):
-        return super().__init__(
-            result_types=[IndexType()], attributes={"dimension": dim}
-        )
+        super().__init__(result_types=[IndexType()], properties={"dimension": dim})
 
 
 @irdl_op_definition
 class DeallocOp(IRDLOperation):
     name = "gpu.dealloc"
 
-    asyncDependencies: VarOperand = var_operand_def(AsyncTokenType)
-    buffer: Operand = operand_def(memref.MemRefType)
+    asyncDependencies = var_operand_def(AsyncTokenType)
+    buffer = operand_def(memref.MemRefType)
 
     irdl_options = [AttrSizedOperandSegments()]
 
-    asyncToken: OptOpResult = opt_result_def(AsyncTokenType)
+    asyncToken = opt_result_def(AsyncTokenType)
 
     def __init__(
         self,
@@ -311,7 +302,7 @@ class DeallocOp(IRDLOperation):
         async_dependencies: Sequence[SSAValue | Operation] | None = None,
         is_async: bool = False,
     ):
-        return super().__init__(
+        super().__init__(
             operands=[async_dependencies, buffer],
             result_types=[[AsyncTokenType()] if is_async else []],
         )
@@ -321,13 +312,13 @@ class DeallocOp(IRDLOperation):
 class MemcpyOp(IRDLOperation):
     name = "gpu.memcpy"
 
-    asyncDependencies: VarOperand = var_operand_def(AsyncTokenType)
-    dst: Operand = operand_def(memref.MemRefType)
-    src: Operand = operand_def(memref.MemRefType)
+    asyncDependencies = var_operand_def(AsyncTokenType)
+    dst = operand_def(memref.MemRefType)
+    src = operand_def(memref.MemRefType)
 
     irdl_options = [AttrSizedOperandSegments()]
 
-    asyncToken: OptOpResult = opt_result_def(AsyncTokenType)
+    asyncToken = opt_result_def(AsyncTokenType)
 
     def __init__(
         self,
@@ -336,7 +327,7 @@ class MemcpyOp(IRDLOperation):
         async_dependencies: Sequence[SSAValue | Operation] | None = None,
         is_async: bool = False,
     ):
-        return super().__init__(
+        super().__init__(
             operands=[async_dependencies, destination, source],
             result_types=[[AsyncTokenType()] if is_async else []],
         )
@@ -349,55 +340,40 @@ class MemcpyOp(IRDLOperation):
             )
 
 
-@irdl_op_definition
-class ModuleEndOp(IRDLOperation):
-    name = "gpu.module_end"
-
-    # TODO circular dependency disallows this set of traits
-    # tracked by gh issues https://github.com/xdslproject/xdsl/issues/1218
-    # traits = frozenset([HasParent(ModuleOp), IsTerminator()])
-    traits = frozenset([IsTerminator()])
-
-    def __init__(self):
-        return super().__init__()
-
-
 @irdl_op_definition
 class ModuleOp(IRDLOperation):
     name = "gpu.module"
 
-    body: Region = region_def("single_block")
-    sym_name: StringAttr = attr_def(StringAttr)
+    body = region_def("single_block")
+    sym_name = prop_def(SymbolNameConstraint())
 
-    traits = frozenset(
-        [
-            IsolatedFromAbove(),
-            SingleBlockImplicitTerminator(ModuleEndOp),
-            SymbolOpInterface(),
-            SymbolTable(),
-        ]
+    traits = traits_def(
+        IsolatedFromAbove(),
+        NoTerminator(),
+        SymbolOpInterface(),
+        SymbolTable(),
     )
 
-    def __init__(self, name: SymbolRefAttr, ops: Sequence[Operation]):
-        super().__init__(attributes={"sym_name": name}, regions=[ops])
+    def __init__(self, name: SymbolRefAttr, body: Region):
+        super().__init__(properties={"sym_name": name}, regions=[body])
 
 
 @irdl_op_definition
 class FuncOp(IRDLOperation):
     name = "gpu.func"
 
-    body: Region = region_def()
-    sym_name: StringAttr = attr_def(StringAttr)
-    function_type: FunctionType = attr_def(FunctionType)
-    kernel: UnitAttr | None = opt_attr_def(UnitAttr)
-    known_block_size: DenseArrayBase | None = opt_attr_def(
-        DenseArrayBase, attr_name="gpu.known_block_size"
+    body = region_def()
+    sym_name = attr_def(SymbolNameConstraint())
+    function_type = prop_def(FunctionType)
+    kernel = opt_prop_def(UnitAttr)
+    known_block_size = opt_attr_def(
+        DenseArrayBase.constr(i32), attr_name="gpu.known_block_size"
     )
-    known_grid_size: DenseArrayBase | None = opt_attr_def(
-        DenseArrayBase, attr_name="gpu.known_grid_size"
+    known_grid_size = opt_attr_def(
+        DenseArrayBase.constr(i32), attr_name="gpu.known_grid_size"
     )
 
-    traits = frozenset([IsolatedFromAbove(), HasParent(ModuleOp), SymbolOpInterface()])
+    traits = traits_def(IsolatedFromAbove(), HasParent(ModuleOp), SymbolOpInterface())
 
     def __init__(
         self,
@@ -405,34 +381,34 @@ class FuncOp(IRDLOperation):
         function_type: FunctionType | tuple[Sequence[Attribute], Sequence[Attribute]],
         region: Region | type[Region.DEFAULT] = Region.DEFAULT,
         kernel: bool | None = None,
-        knwown_block_size: Sequence[int] | None = None,
-        knwown_grid_size: Sequence[int] | None = None,
+        known_block_size: Sequence[int] | None = None,
+        known_grid_size: Sequence[int] | None = None,
     ):
         if isinstance(function_type, tuple):
             inputs, outputs = function_type
             function_type = FunctionType.from_lists(inputs, outputs)
         if not isinstance(region, Region):
             region = Region(Block(arg_types=function_type.inputs))
-        attributes: dict[str, Attribute | None] = {
-            "sym_name": StringAttr(name),
+        attributes: dict[str, Attribute | None] = {"sym_name": StringAttr(name)}
+        properties: dict[str, Attribute | None] = {
             "function_type": function_type,
         }
-        if knwown_block_size is not None:
-            attributes["gpu.known_block_size"] = ArrayAttr(
-                IntegerAttr(i, i32) for i in knwown_block_size
+        if known_block_size is not None:
+            attributes["gpu.known_block_size"] = DenseArrayBase.from_list(
+                i32, known_block_size
             )
-        if knwown_grid_size is not None:
-            attributes["gpu.known_grid_size"] = ArrayAttr(
-                IntegerAttr(i, i32) for i in knwown_grid_size
+        if known_grid_size is not None:
+            attributes["gpu.known_grid_size"] = DenseArrayBase.from_list(
+                i32, known_grid_size
             )
         if kernel:
-            attributes["kernel"] = UnitAttr()
-        super().__init__(attributes=attributes, regions=[region])
+            properties["kernel"] = UnitAttr()
+        super().__init__(properties=properties, attributes=attributes, regions=[region])
 
     def verify_(self):
         entry_block: Block = self.body.blocks[0]
         function_inputs = self.function_type.inputs.data
-        block_arg_types = tuple(a.type for a in entry_block.args)
+        block_arg_types = entry_block.arg_types
         if function_inputs != block_arg_types:
             raise VerifyException(
                 "Expected first entry block arguments to have the same types as the "
@@ -445,25 +421,21 @@ class FuncOp(IRDLOperation):
 @irdl_op_definition
 class GlobalIdOp(IRDLOperation):
     name = "gpu.global_id"
-    dimension: DimensionAttr = attr_def(DimensionAttr)
-    result: OpResult = result_def(IndexType)
+    dimension = prop_def(DimensionAttr)
+    result = result_def(IndexType)
 
     def __init__(self, dim: DimensionAttr):
-        return super().__init__(
-            result_types=[IndexType()], attributes={"dimension": dim}
-        )
+        super().__init__(result_types=[IndexType()], properties={"dimension": dim})
 
 
 @irdl_op_definition
 class GridDimOp(IRDLOperation):
     name = "gpu.grid_dim"
-    dimension: DimensionAttr = attr_def(DimensionAttr)
-    result: OpResult = result_def(IndexType)
+    dimension = prop_def(DimensionAttr)
+    result = result_def(IndexType)
 
     def __init__(self, dim: DimensionAttr):
-        return super().__init__(
-            result_types=[IndexType()], attributes={"dimension": dim}
-        )
+        super().__init__(result_types=[IndexType()], properties={"dimension": dim})
 
 
 @irdl_op_definition
@@ -480,10 +452,10 @@ class HostRegisterOp(IRDLOperation):
 
     name = "gpu.host_register"
 
-    value: Operand = operand_def(memref.UnrankedMemrefType)
+    value = operand_def(memref.UnrankedMemRefType)
 
     def __init__(self, memref: SSAValue | Operation):
-        return super().__init__(operands=[SSAValue.get(memref)])
+        super().__init__(operands=[SSAValue.get(memref)])
 
 
 @irdl_op_definition
@@ -494,41 +466,45 @@ class HostUnregisterOp(IRDLOperation):
 
     name = "gpu.host_unregister"
 
-    value: Operand = operand_def(memref.UnrankedMemrefType)
+    value = operand_def(memref.UnrankedMemRefType)
 
     def __init__(self, memref: SSAValue | Operation):
-        return super().__init__(operands=[SSAValue.get(memref)])
+        super().__init__(operands=[SSAValue.get(memref)])
 
 
 @irdl_op_definition
 class LaneIdOp(IRDLOperation):
     name = "gpu.lane_id"
-    result: OpResult = result_def(IndexType)
+    result = result_def(IndexType)
 
     def __init__(self):
-        return super().__init__(result_types=[IndexType()])
+        super().__init__(result_types=[IndexType()])
 
 
 @irdl_op_definition
 class LaunchOp(IRDLOperation):
     name = "gpu.launch"
-    asyncDependencies: VarOperand = var_operand_def(AsyncTokenType)
-    gridSizeX: Operand = operand_def(IndexType)
-    gridSizeY: Operand = operand_def(IndexType)
-    gridSizeZ: Operand = operand_def(IndexType)
-    blockSizeX: Operand = operand_def(IndexType)
-    blockSizeY: Operand = operand_def(IndexType)
-    blockSizeZ: Operand = operand_def(IndexType)
-    dynamicSharedMemorySize: OptOperand = opt_operand_def(i32)
-    asyncToken: OptOpResult = opt_result_def(AsyncTokenType)
-    body: Region = region_def()
-    irdl_options = [AttrSizedOperandSegments()]
+    asyncDependencies = var_operand_def(AsyncTokenType)
+    gridSizeX = operand_def(IndexType)
+    gridSizeY = operand_def(IndexType)
+    gridSizeZ = operand_def(IndexType)
+    blockSizeX = operand_def(IndexType)
+    blockSizeY = operand_def(IndexType)
+    blockSizeZ = operand_def(IndexType)
+    clusterSizeX = opt_operand_def(IndexType)
+    clusterSizeY = opt_operand_def(IndexType)
+    clusterSizeZ = opt_operand_def(IndexType)
+    dynamicSharedMemorySize = opt_operand_def(i32)
+    asyncToken = opt_result_def(AsyncTokenType)
+    body = region_def()
+    irdl_options = [AttrSizedOperandSegments(as_property=True)]
 
     def __init__(
         self,
         body: Region,
         gridSize: Sequence[SSAValue | Operation],
         blockSize: Sequence[SSAValue | Operation],
+        clusterSize: Sequence[SSAValue | Operation] = [],
         async_launch: bool = False,
         asyncDependencies: Sequence[SSAValue | Operation] | None = None,
         dynamicSharedMemorySize: SSAValue | Operation | None = None,
@@ -537,20 +513,32 @@ class LaunchOp(IRDLOperation):
             raise ValueError(f"LaunchOp must have 3 gridSizes, got {len(gridSize)}")
         if len(blockSize) != 3:
             raise ValueError(f"LaunchOp must have 3 blockSizes, got {len(blockSize)}")
+        if len(clusterSize) != 3 and len(clusterSize) != 0:
+            raise ValueError(
+                f"LaunchOp must have 0 or 3 clusterSizes, got {len(clusterSize)}"
+            )
         operands = [
-            []
-            if asyncDependencies is None
-            else [SSAValue.get(a) for a in asyncDependencies]
+            (
+                []
+                if asyncDependencies is None
+                else [SSAValue.get(a) for a in asyncDependencies]
+            )
         ]
 
         operands += [SSAValue.get(gs) for gs in gridSize]
         operands += [SSAValue.get(bs) for bs in blockSize]
+        if clusterSize:
+            operands += [(SSAValue.get(cs),) for cs in clusterSize]
+        else:
+            operands += [(), (), ()]
         operands += [
-            []
-            if dynamicSharedMemorySize is None
-            else [SSAValue.get(dynamicSharedMemorySize)]
+            (
+                []
+                if dynamicSharedMemorySize is None
+                else [SSAValue.get(dynamicSharedMemorySize)]
+            )
         ]
-        return super().__init__(
+        super().__init__(
             operands=operands,
             result_types=[[AsyncTokenType()] if async_launch else []],
             regions=[body],
@@ -559,9 +547,8 @@ class LaunchOp(IRDLOperation):
     def verify_(self) -> None:
         if not any(b.ops for b in self.body.blocks):
             raise VerifyException("gpu.launch requires a non-empty body.")
-        body_args = self.body.blocks[0].args
-        args_type = [a.type for a in body_args]
-        if args_type != [IndexType()] * 12:
+        args_type = self.body.blocks[0].arg_types
+        if args_type != (IndexType(),) * 12:
             raise VerifyException(
                 f"Expected [12 x {str(IndexType())}], got {[str(t) for t in args_type]}. "
                 "gpu.launch's body arguments are 12 index arguments, with 3 block "
@@ -601,27 +588,32 @@ class LaunchFuncOp(IRDLOperation):
     """
 
     name = "gpu.launch_func"
-    asyncDependencies: VarOperand = var_operand_def(AsyncTokenType)
-    gridSizeX: Operand = operand_def(IndexType)
-    gridSizeY: Operand = operand_def(IndexType)
-    gridSizeZ: Operand = operand_def(IndexType)
-    blockSizeX: Operand = operand_def(IndexType)
-    blockSizeY: Operand = operand_def(IndexType)
-    blockSizeZ: Operand = operand_def(IndexType)
-    dynamicSharedMemorySize: OptOperand = opt_operand_def(i32)
-    kernelOperands: VarOperand = var_operand_def()
+    asyncDependencies = var_operand_def(AsyncTokenType)
+    gridSizeX = operand_def(AnyOf((IndexType, i32, i64)))
+    gridSizeY = operand_def(AnyOf((IndexType, i32, i64)))
+    gridSizeZ = operand_def(AnyOf((IndexType, i32, i64)))
+    blockSizeX = operand_def(AnyOf((IndexType, i32, i64)))
+    blockSizeY = operand_def(AnyOf((IndexType, i32, i64)))
+    blockSizeZ = operand_def(AnyOf((IndexType, i32, i64)))
+    clusterSizeX = opt_operand_def(AnyOf((IndexType, i32, i64)))
+    clusterSizeY = opt_operand_def(AnyOf((IndexType, i32, i64)))
+    clusterSizeZ = opt_operand_def(AnyOf((IndexType, i32, i64)))
+    dynamicSharedMemorySize = opt_operand_def(i32)
+    kernelOperands = var_operand_def()
+    asyncObject = opt_operand_def()
 
-    asyncToken: OptOpResult = opt_result_def(AsyncTokenType)
+    asyncToken = opt_result_def(AsyncTokenType)
 
-    kernel: SymbolRefAttr = attr_def(SymbolRefAttr)
+    kernel = prop_def(SymbolRefAttr)
 
-    irdl_options = [AttrSizedOperandSegments()]
+    irdl_options = [AttrSizedOperandSegments(as_property=True)]
 
     def __init__(
         self,
         func: SymbolRefAttr,
         gridSize: Sequence[SSAValue | Operation],
         blockSize: Sequence[SSAValue | Operation],
+        clusterSize: Sequence[SSAValue | Operation] | None = None,
         kernelOperands: Sequence[SSAValue | Operation] | None = None,
         async_launch: bool = False,
         asyncDependencies: Sequence[SSAValue | Operation] | None = None,
@@ -631,105 +623,132 @@ class LaunchFuncOp(IRDLOperation):
             raise ValueError(f"LaunchOp must have 3 gridSizes, got {len(gridSize)}")
         if len(blockSize) != 3:
             raise ValueError(f"LaunchOp must have 3 blockSizes, got {len(blockSize)}")
+        clusterSizeOperands: Sequence[
+            SSAValue | Operation | Sequence[SSAValue | Operation]
+        ]
+        if clusterSize is None:
+            clusterSizeOperands = [[], [], []]
+        else:
+            clusterSizeOperands = clusterSize
+        if len(clusterSizeOperands) != 3:
+            raise ValueError(
+                f"LaunchFuncOp must have 3 cluterSizes if any, got {len(clusterSizeOperands)}"
+            )
 
-        return super().__init__(
+        super().__init__(
             operands=[
                 asyncDependencies,
                 *gridSize,
                 *blockSize,
+                *clusterSizeOperands,
                 dynamicSharedMemorySize,
                 kernelOperands,
+                [],
             ],
             result_types=[[AsyncTokenType()] if async_launch else []],
-            attributes={"kernel": func},
+            properties={"kernel": func},
         )
 
 
 @irdl_op_definition
 class NumSubgroupsOp(IRDLOperation):
     name = "gpu.num_subgroups"
-    result: OpResult = result_def(IndexType)
+    result = result_def(IndexType)
 
     def __init__(self):
-        return super().__init__(result_types=[IndexType()])
+        super().__init__(result_types=[IndexType()])
 
 
 @irdl_op_definition
 class ReturnOp(IRDLOperation):
     name = "gpu.return"
 
-    args: VarOperand = var_operand_def()
+    args = var_operand_def()
 
-    traits = frozenset([IsTerminator(), HasParent(FuncOp)])
+    traits = traits_def(IsTerminator(), HasParent(FuncOp))
 
     def __init__(self, operands: Sequence[SSAValue | Operation]):
-        return super().__init__([operands])
+        super().__init__(operands=[operands])
 
 
 @irdl_op_definition
 class SetDefaultDeviceOp(IRDLOperation):
     name = "gpu.set_default_device"
-    devIndex: Operand = operand_def(i32)
+    devIndex = operand_def(i32)
 
     def __init__(self, devIndex: SSAValue | Operation):
-        return super().__init__(operands=[SSAValue.get(devIndex)])
+        super().__init__(operands=[SSAValue.get(devIndex)])
 
 
 @irdl_op_definition
 class SubgroupIdOp(IRDLOperation):
     name = "gpu.subgroup_id"
-    result: OpResult = result_def(IndexType)
+    result = result_def(IndexType)
 
     def __init__(self):
-        return super().__init__(result_types=[IndexType()])
+        super().__init__(result_types=[IndexType()])
 
 
 @irdl_op_definition
 class SubgroupSizeOp(IRDLOperation):
     name = "gpu.subgroup_size"
-    result: OpResult = result_def(IndexType)
+    result = result_def(IndexType)
 
     def __init__(self):
-        return super().__init__(result_types=[IndexType()])
+        super().__init__(result_types=[IndexType()])
 
 
 @irdl_op_definition
 class TerminatorOp(IRDLOperation):
     name = "gpu.terminator"
 
-    traits = frozenset([HasParent(LaunchOp), IsTerminator()])
+    traits = traits_def(HasParent(LaunchOp), IsTerminator())
 
     def __init__(self):
-        return super().__init__()
+        super().__init__()
 
 
 @irdl_op_definition
 class ThreadIdOp(IRDLOperation):
     name = "gpu.thread_id"
-    dimension: DimensionAttr = attr_def(DimensionAttr)
-    result: OpResult = result_def(IndexType)
+    dimension = prop_def(DimensionAttr)
+    result = result_def(IndexType)
 
     def __init__(self, dim: DimensionAttr):
-        return super().__init__(
-            result_types=[IndexType()], attributes={"dimension": dim}
+        super().__init__(result_types=[IndexType()], properties={"dimension": dim})
+
+
+@irdl_op_definition
+class WaitOp(IRDLOperation):
+    name = "gpu.wait"
+    asyncDependencies = var_operand_def(AsyncTokenType)
+    asyncToken = opt_result_def(AsyncTokenType)
+
+    def __init__(
+        self,
+        async_dependencies: Sequence[SSAValue | Operation] | None = None,
+    ):
+        super().__init__(
+            operands=[async_dependencies],
+            result_types=[[AsyncTokenType()]],
         )
 
 
 @irdl_op_definition
 class YieldOp(IRDLOperation):
     name = "gpu.yield"
-    values: VarOperand = var_operand_def(Attribute)
+    values = var_operand_def(Attribute)
 
     def __init__(self, operands: Sequence[SSAValue | Operation]):
-        return super().__init__([operands])
+        super().__init__(operands=[operands])
 
-    traits = frozenset([IsTerminator()])
+    traits = traits_def(IsTerminator())
 
     def verify_(self) -> None:
         op = self.parent_op()
         if op is not None:
-            yield_type = [o.type for o in self.values]
-            result_type = [r.type for r in op.results]
+            yield_type = self.values.types
+            result_type = op.result_types
             if yield_type != result_type:
                 raise VerifyException(
                     f"Expected {[str(t) for t in result_type]}, got {[str(t) for t in yield_type]}. The gpu.yield values "
@@ -737,11 +756,8 @@ class YieldOp(IRDLOperation):
                 )
 
 
-# _GPUAttr has to be registered instead of DimensionAttr and AllReduceOperationAttr here.
-# This is a hack to fit MLIR's syntax in xDSL's way of parsing attributes, without making GPU builtin.
-# Hopefully MLIR will parse it in a more xDSL-friendly way soon, so all that can be factored in proper xDSL
-# atrributes.
 GPU = Dialect(
+    "gpu",
     [
         AllocOp,
         AllReduceOp,
@@ -759,7 +775,6 @@ GPU = Dialect(
         LaunchFuncOp,
         MemcpyOp,
         ModuleOp,
-        ModuleEndOp,
         NumSubgroupsOp,
         ReturnOp,
         SetDefaultDeviceOp,
@@ -767,7 +782,14 @@ GPU = Dialect(
         SubgroupSizeOp,
         TerminatorOp,
         ThreadIdOp,
+        WaitOp,
         YieldOp,
     ],
-    [_GPUAttr],
+    [
+        AsyncTokenType,
+        AllReduceOpAttr,
+        DimensionAttr,
+        ProcessorAttr,
+        LoopDimMapAttr,
+    ],
 )
