@@ -1,17 +1,25 @@
 from __future__ import annotations
 
 from abc import ABC
+from collections.abc import Sequence
 from enum import Enum
-from typing import Generic, Sequence, TypeVar, cast
+from typing import Generic
+
+from typing_extensions import TypeVar
 
 from xdsl.dialects import llvm
-from xdsl.dialects.builtin import AnyFloat, IntegerType, Signedness, StringAttr, i32
-from xdsl.dialects.memref import MemRefType
+from xdsl.dialects.builtin import (
+    AnyFloat,
+    IntegerType,
+    MemRefType,
+    Signedness,
+    StringAttr,
+    i32,
+)
 from xdsl.ir import (
     Attribute,
     Dialect,
     Operation,
-    OpResult,
     ParametrizedAttribute,
     SSAValue,
     TypeAttribute,
@@ -19,10 +27,8 @@ from xdsl.ir import (
 from xdsl.irdl import (
     IRDLOperation,
     Operand,
-    OptOperand,
-    OptOpResult,
-    ParameterDef,
     attr_def,
+    base,
     irdl_attr_definition,
     irdl_op_definition,
     operand_def,
@@ -48,7 +54,7 @@ class OperationType(ParametrizedAttribute, TypeAttribute):
 
     name = "mpi.operation"
 
-    op_str: ParameterDef[StringAttr]
+    op_str: StringAttr
 
 
 class MpiOp:
@@ -56,20 +62,20 @@ class MpiOp:
     A collection of MPI_Op types used for
     """
 
-    MPI_MAX = OperationType([StringAttr("MPI_MAX")])
-    MPI_MIN = OperationType([StringAttr("MPI_MIN")])
-    MPI_SUM = OperationType([StringAttr("MPI_SUM")])
-    MPI_PROD = OperationType([StringAttr("MPI_PROD")])
-    MPI_LAND = OperationType([StringAttr("MPI_LAND")])
-    MPI_BAND = OperationType([StringAttr("MPI_BAND")])
-    MPI_LOR = OperationType([StringAttr("MPI_LOR")])
-    MPI_BOR = OperationType([StringAttr("MPI_BOR")])
-    MPI_LXOR = OperationType([StringAttr("MPI_LXOR")])
-    MPI_BXOR = OperationType([StringAttr("MPI_BXOR")])
-    MPI_MINLOC = OperationType([StringAttr("MPI_MINLOC")])
-    MPI_MAXLOC = OperationType([StringAttr("MPI_MAXLOC")])
-    MPI_REPLACE = OperationType([StringAttr("MPI_REPLACE")])
-    MPI_NO_OP = OperationType([StringAttr("MPI_NO_OP")])
+    MPI_MAX = OperationType(StringAttr("MPI_MAX"))
+    MPI_MIN = OperationType(StringAttr("MPI_MIN"))
+    MPI_SUM = OperationType(StringAttr("MPI_SUM"))
+    MPI_PROD = OperationType(StringAttr("MPI_PROD"))
+    MPI_LAND = OperationType(StringAttr("MPI_LAND"))
+    MPI_BAND = OperationType(StringAttr("MPI_BAND"))
+    MPI_LOR = OperationType(StringAttr("MPI_LOR"))
+    MPI_BOR = OperationType(StringAttr("MPI_BOR"))
+    MPI_LXOR = OperationType(StringAttr("MPI_LXOR"))
+    MPI_BXOR = OperationType(StringAttr("MPI_BXOR"))
+    MPI_MINLOC = OperationType(StringAttr("MPI_MINLOC"))
+    MPI_MAXLOC = OperationType(StringAttr("MPI_MAXLOC"))
+    MPI_REPLACE = OperationType(StringAttr("MPI_REPLACE"))
+    MPI_NO_OP = OperationType(StringAttr("MPI_NO_OP"))
 
 
 @irdl_attr_definition
@@ -104,21 +110,18 @@ class DataType(ParametrizedAttribute, TypeAttribute):
 
 
 VectorWrappable = RequestType | StatusType | DataType
-_VectorT = TypeVar("_VectorT", bound=VectorWrappable)
+VectorWrappableConstr = base(RequestType) | base(StatusType) | base(DataType)
+_VectorT = TypeVar("_VectorT", bound=VectorWrappable, default=VectorWrappable)
 
 
 @irdl_attr_definition
-class VectorType(Generic[_VectorT], ParametrizedAttribute, TypeAttribute):
+class VectorType(ParametrizedAttribute, TypeAttribute, Generic[_VectorT]):
     """
     This type holds multiple MPI types
     """
 
     name = "mpi.vector"
-    wrapped_type: ParameterDef[_VectorT]
-
-    @staticmethod
-    def of(dtype: type[_VectorT]) -> VectorType[_VectorT]:
-        return VectorType([dtype([])])
+    wrapped_type: _VectorT
 
 
 class StatusTypeField(Enum):
@@ -140,10 +143,11 @@ class MPIBaseOp(IRDLOperation, ABC):
 
 
 @irdl_op_definition
-class Reduce(MPIBaseOp):
+class ReduceOp(MPIBaseOp):
     """
     This wraps the MPI_Reduce function (blocking reduction)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Reduce.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Reduce.html).
 
     ## The MPI_Reduce Function Docs:
 
@@ -166,15 +170,15 @@ class Reduce(MPIBaseOp):
 
     name = "mpi.reduce"
 
-    send_buffer: Operand = operand_def(Attribute)
-    recv_buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    operationtype: OperationType = attr_def(OperationType)
-    root: Operand = operand_def(i32)
+    send_buffer = operand_def(Attribute)
+    recv_buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    operationtype = attr_def(OperationType)
+    root = operand_def(i32)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         send_buffer: SSAValue | Operation,
         recv_buffer: SSAValue | Operation,
         count: SSAValue | Operation,
@@ -182,7 +186,7 @@ class Reduce(MPIBaseOp):
         operationtype: OperationType,
         root: SSAValue | Operation,
     ):
-        return Reduce.build(
+        return super().__init__(
             operands=[send_buffer, recv_buffer, count, datatype, root],
             attributes={"operationtype": operationtype},
             result_types=[],
@@ -190,10 +194,11 @@ class Reduce(MPIBaseOp):
 
 
 @irdl_op_definition
-class Allreduce(MPIBaseOp):
+class AllreduceOp(MPIBaseOp):
     """
     This wraps the MPI_Allreduce function (blocking all reduction)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Allreduce.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Allreduce.html).
 
     ## The MPI_Allreduce Function Docs:
 
@@ -215,14 +220,14 @@ class Allreduce(MPIBaseOp):
 
     name = "mpi.allreduce"
 
-    send_buffer: OptOperand = opt_operand_def(Attribute)
-    recv_buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    operationtype: OperationType = attr_def(OperationType)
+    send_buffer = opt_operand_def(Attribute)
+    recv_buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    operationtype = attr_def(OperationType)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         send_buffer: SSAValue | Operation | None,
         recv_buffer: SSAValue | Operation,
         count: SSAValue | Operation,
@@ -232,12 +237,13 @@ class Allreduce(MPIBaseOp):
         operands_to_add: Sequence[
             SSAValue | Operation | Sequence[SSAValue | Operation]
         ] = []
+
         if send_buffer is None:
             operands_to_add = [[], recv_buffer, count, datatype]
         else:
             operands_to_add = [[send_buffer], recv_buffer, count, datatype]
 
-        return Allreduce.build(
+        return super().__init__(
             operands=operands_to_add,
             attributes={"operationtype": operationtype},
             result_types=[],
@@ -245,10 +251,11 @@ class Allreduce(MPIBaseOp):
 
 
 @irdl_op_definition
-class Bcast(MPIBaseOp):
+class BcastOp(MPIBaseOp):
     """
     This wraps the MPI_Bcast function (blocking broadcast)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Bcast.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Bcast.html).
 
     ## The MPI_Bcast Function Docs:
 
@@ -269,29 +276,30 @@ class Bcast(MPIBaseOp):
 
     name = "mpi.bcast"
 
-    buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    root: Operand = operand_def(i32)
+    buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    root = operand_def(i32)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         buffer: SSAValue | Operation,
         count: SSAValue | Operation,
         datatype: SSAValue | Operation,
         root: SSAValue | Operation,
     ):
-        return Bcast.build(
+        return super().__init__(
             operands=[buffer, count, datatype, root],
             result_types=[],
         )
 
 
 @irdl_op_definition
-class Isend(MPIBaseOp):
+class IsendOp(MPIBaseOp):
     """
     This wraps the MPI_Isend function (nonblocking send)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Isend.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Isend.html).
 
     ## The MPI_Isend Function Docs:
 
@@ -313,15 +321,15 @@ class Isend(MPIBaseOp):
 
     name = "mpi.isend"
 
-    buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    dest: Operand = operand_def(i32)
-    tag: Operand = operand_def(i32)
-    request: Operand = operand_def(RequestType)
+    buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    dest = operand_def(i32)
+    tag = operand_def(i32)
+    request = operand_def(RequestType)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         buffer: SSAValue | Operation,
         count: SSAValue | Operation,
         datatype: SSAValue | Operation,
@@ -329,17 +337,18 @@ class Isend(MPIBaseOp):
         tag: SSAValue | Operation,
         request: SSAValue | Operation,
     ):
-        return Isend.build(
+        return super().__init__(
             operands=[buffer, count, datatype, dest, tag, request],
             result_types=[],
         )
 
 
 @irdl_op_definition
-class Send(MPIBaseOp):
+class SendOp(MPIBaseOp):
     """
     This wraps the MPI_Send function (blocking send)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Send.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Send.html).
 
     ## The MPI_Send Function Docs:
 
@@ -361,30 +370,31 @@ class Send(MPIBaseOp):
 
     name = "mpi.send"
 
-    buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    dest: Operand = operand_def(i32)
-    tag: Operand = operand_def(i32)
+    buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    dest = operand_def(i32)
+    tag = operand_def(i32)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         buffer: SSAValue | Operation,
         count: SSAValue | Operation,
         datatype: SSAValue | Operation,
         dest: SSAValue | Operation,
         tag: SSAValue | Operation,
-    ) -> Send:
-        return Send.build(
+    ):
+        return super().__init__(
             operands=[buffer, count, datatype, dest, tag], result_types=[]
         )
 
 
 @irdl_op_definition
-class Irecv(MPIBaseOp):
+class IrecvOp(MPIBaseOp):
     """
     This wraps the MPI_Irecv function (nonblocking receive).
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Irecv.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Irecv.html).
 
     ## The MPI_Irecv Function Docs:
 
@@ -407,15 +417,15 @@ class Irecv(MPIBaseOp):
 
     name = "mpi.irecv"
 
-    buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    source: Operand = operand_def(i32)
-    tag: Operand = operand_def(i32)
-    request: Operand = operand_def(RequestType)
+    buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    source = operand_def(i32)
+    tag = operand_def(i32)
+    request = operand_def(RequestType)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         buffer: SSAValue | Operation,
         count: SSAValue | Operation,
         datatype: SSAValue | Operation,
@@ -423,17 +433,18 @@ class Irecv(MPIBaseOp):
         tag: SSAValue | Operation,
         request: SSAValue | Operation,
     ):
-        return Irecv.build(
+        return super().__init__(
             operands=[buffer, count, datatype, source, tag, request],
             result_types=[],
         )
 
 
 @irdl_op_definition
-class Recv(MPIBaseOp):
+class RecvOp(MPIBaseOp):
     """
     This wraps the MPI_Recv function (blocking receive).
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Recv.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Recv.html).
 
     ## The MPI_Recv Function Docs:
 
@@ -456,16 +467,16 @@ class Recv(MPIBaseOp):
 
     name = "mpi.recv"
 
-    buffer: Operand = operand_def(Attribute)
-    count: Operand = operand_def(i32)
-    datatype: Operand = operand_def(DataType)
-    source: Operand = operand_def(i32)
-    tag: Operand = operand_def(i32)
+    buffer = operand_def(Attribute)
+    count = operand_def(i32)
+    datatype = operand_def(DataType)
+    source = operand_def(i32)
+    tag = operand_def(i32)
 
-    status: OptOpResult = opt_result_def(StatusType)
+    status = opt_result_def(StatusType)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         buffer: SSAValue | Operation,
         count: SSAValue | Operation,
         datatype: SSAValue | Operation,
@@ -473,17 +484,18 @@ class Recv(MPIBaseOp):
         tag: SSAValue | Operation,
         ignore_status: bool = True,
     ):
-        return Recv.build(
+        return super().__init__(
             operands=[buffer, count, datatype, source, tag],
             result_types=[[]] if ignore_status else [[StatusType()]],
         )
 
 
 @irdl_op_definition
-class Test(MPIBaseOp):
+class TestOp(MPIBaseOp):
     """
     Class for wrapping the MPI_Test function (test for completion of request)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Test.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Test.html).
 
     ## The MPI_Test Function Docs:
 
@@ -496,21 +508,21 @@ class Test(MPIBaseOp):
 
     name = "mpi.test"
 
-    request: Operand = operand_def(RequestType)
+    request = operand_def(RequestType)
 
-    flag: OpResult = result_def(t_bool)
-    status: OpResult = result_def(StatusType)
+    flag = result_def(t_bool)
+    status = result_def(StatusType)
 
-    @staticmethod
-    def get(request: Operand):
-        return Test.build(operands=[request], result_types=[t_bool, StatusType()])
+    def __init__(self, request: Operand):
+        return super().__init__(operands=[request], result_types=[t_bool, StatusType()])
 
 
 @irdl_op_definition
-class Wait(MPIBaseOp):
+class WaitOp(MPIBaseOp):
     """
     Class for wrapping the MPI_Wait function (blocking wait for request)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Wait.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Wait.html).
 
     ## The MPI_Test Function Docs:
 
@@ -522,23 +534,23 @@ class Wait(MPIBaseOp):
 
     name = "mpi.wait"
 
-    request: Operand = operand_def(RequestType)
-    status: OptOpResult = opt_result_def(StatusType)
+    request = operand_def(RequestType)
+    status = opt_result_def(StatusType)
 
-    @staticmethod
-    def get(request: Operand, ignore_status: bool = True):
+    def __init__(self, request: Operand, ignore_status: bool = True):
         result_types: list[list[Attribute]] = [[StatusType()]]
         if ignore_status:
             result_types = [[]]
 
-        return Wait.build(operands=[request], result_types=result_types)
+        return super().__init__(operands=[request], result_types=result_types)
 
 
 @irdl_op_definition
-class Waitall(MPIBaseOp):
+class WaitallOp(MPIBaseOp):
     """
     Class for wrapping the MPI_Waitall function (blocking wait for requests)
-    https://www.mpich.org/static/docs/v4.1/www3/MPI_Waitall.html
+
+    See external [documentation](https://www.mpich.org/static/docs/v4.1/www3/MPI_Waitall.html).
 
     ## The MPI_Test Function Docs:
 
@@ -552,21 +564,20 @@ class Waitall(MPIBaseOp):
 
     name = "mpi.waitall"
 
-    requests: Operand = operand_def(VectorType[RequestType])
-    count: Operand = operand_def(i32)
-    statuses: OptOpResult = opt_result_def(VectorType[StatusType])
+    requests = operand_def(VectorType[RequestType])
+    count = operand_def(i32)
+    statuses = opt_result_def(VectorType[StatusType])
 
-    @staticmethod
-    def get(requests: Operand, count: Operand, ignore_status: bool = True):
-        result_types: list[list[Attribute]] = [[VectorType.of(StatusType)]]
+    def __init__(self, requests: Operand, count: Operand, ignore_status: bool = True):
+        result_types: list[list[Attribute]] = [[VectorType(StatusType())]]
         if ignore_status:
             result_types = [[]]
 
-        return Waitall.build(operands=[requests, count], result_types=result_types)
+        return super().__init__(operands=[requests, count], result_types=result_types)
 
 
 @irdl_op_definition
-class GetStatusField(MPIBaseOp):
+class GetStatusFieldOp(MPIBaseOp):
     """
     Accessors for the MPI_Status struct
 
@@ -580,15 +591,14 @@ class GetStatusField(MPIBaseOp):
 
     name = "mpi.status.get"
 
-    status: Operand = operand_def(StatusType)
+    status = operand_def(StatusType)
 
-    field: StringAttr = attr_def(StringAttr)
+    field = attr_def(StringAttr)
 
-    result: OpResult = result_def(i32)
+    result = result_def(i32)
 
-    @staticmethod
-    def get(status_obj: Operand, field: StatusTypeField):
-        return GetStatusField.build(
+    def __init__(self, status_obj: Operand, field: StatusTypeField):
+        return super().__init__(
             operands=[status_obj],
             attributes={"field": StringAttr(field.value)},
             result_types=[i32],
@@ -596,7 +606,7 @@ class GetStatusField(MPIBaseOp):
 
 
 @irdl_op_definition
-class CommRank(MPIBaseOp):
+class CommRankOp(MPIBaseOp):
     """
     Represents the MPI_Comm_size(MPI_Comm comm, int *rank) function call which returns
     the rank of the communicator
@@ -606,15 +616,14 @@ class CommRank(MPIBaseOp):
 
     name = "mpi.comm.rank"
 
-    rank: OpResult = result_def(i32)
+    rank = result_def(i32)
 
-    @staticmethod
-    def get():
-        return CommRank.build(result_types=[i32])
+    def __init__(self):
+        return super().__init__(result_types=[i32])
 
 
 @irdl_op_definition
-class CommSize(MPIBaseOp):
+class CommSizeOp(MPIBaseOp):
     """
     Represents the MPI_Comm_size(MPI_Comm comm, int *size) function call which returns
     the size of the communicator
@@ -624,15 +633,14 @@ class CommSize(MPIBaseOp):
 
     name = "mpi.comm.size"
 
-    size: OpResult = result_def(i32)
+    size = result_def(i32)
 
-    @staticmethod
-    def get():
-        return CommSize.build(result_types=[i32])
+    def __init__(self):
+        return super().__init__(result_types=[i32])
 
 
 @irdl_op_definition
-class Init(MPIBaseOp):
+class InitOp(MPIBaseOp):
     """
     This represents a bare MPI_Init call with both args being nullptr
     """
@@ -641,7 +649,7 @@ class Init(MPIBaseOp):
 
 
 @irdl_op_definition
-class Finalize(MPIBaseOp):
+class FinalizeOp(MPIBaseOp):
     """
     This represents an MPI_Finalize call with both args being nullptr
     """
@@ -650,7 +658,7 @@ class Finalize(MPIBaseOp):
 
 
 @irdl_op_definition
-class UnwrapMemrefOp(MPIBaseOp):
+class UnwrapMemRefOp(MPIBaseOp):
     """
     This Op can be used as a helper to get memrefs into MPI calls.
 
@@ -659,21 +667,16 @@ class UnwrapMemrefOp(MPIBaseOp):
 
     name = "mpi.unwrap_memref"
 
-    ref: Operand = operand_def(MemRefType[AnyNumericType])
+    ref = operand_def(MemRefType[AnyNumericType])
 
-    ptr: OpResult = result_def(llvm.LLVMPointerType)
-    len: OpResult = result_def(i32)
-    type: OpResult = result_def(DataType)
+    ptr = result_def(llvm.LLVMPointerType)
+    len = result_def(i32)
+    type = result_def(DataType)
 
-    @staticmethod
-    def get(ref: SSAValue | Operation) -> UnwrapMemrefOp:
-        ssa_val = SSAValue.get(ref)
-        assert isinstance(ssa_val.type, MemRefType)
-        elem_type = cast(MemRefType[AnyNumericType], ssa_val.type).element_type
-
-        return UnwrapMemrefOp.build(
+    def __init__(self, ref: SSAValue | Operation):
+        return super().__init__(
             operands=[ref],
-            result_types=[llvm.LLVMPointerType.typed(elem_type), i32, DataType()],
+            result_types=[llvm.LLVMPointerType.opaque(), i32, DataType()],
         )
 
 
@@ -692,13 +695,12 @@ class GetDtypeOp(MPIBaseOp):
 
     name = "mpi.get_dtype"
 
-    dtype: Attribute = attr_def(Attribute)
+    dtype = attr_def()
 
-    result: OpResult = result_def(DataType)
+    result = result_def(DataType)
 
-    @staticmethod
-    def get(dtype: Attribute):
-        return GetDtypeOp.build(result_types=[DataType()], attributes={"dtype": dtype})
+    def __init__(self, dtype: Attribute):
+        return super().__init__(result_types=[DataType()], attributes={"dtype": dtype})
 
 
 @irdl_op_definition
@@ -715,20 +717,20 @@ class AllocateTypeOp(MPIBaseOp):
 
     name = "mpi.allocate"
 
-    bindc_name: StringAttr | None = opt_attr_def(StringAttr)
-    dtype: VectorWrappable = attr_def(VectorWrappable)
-    count: Operand = operand_def(i32)
+    bindc_name = opt_attr_def(StringAttr)
+    dtype = attr_def(VectorWrappableConstr)
+    count = operand_def(i32)
 
-    result: OpResult = result_def(VectorType)
+    result = result_def(VectorType)
 
-    @staticmethod
-    def get(
+    def __init__(
+        self,
         dtype: type[VectorWrappable],
         count: SSAValue | Operation,
         bindc_name: StringAttr | None = None,
-    ) -> AllocateTypeOp:
-        return AllocateTypeOp.build(
-            result_types=[VectorType.of(dtype)],
+    ):
+        return super().__init__(
+            result_types=[VectorType(dtype())],
             attributes={
                 "dtype": dtype(),
                 "bindc_name": bindc_name,
@@ -746,17 +748,16 @@ class VectorGetOp(MPIBaseOp):
 
     name = "mpi.vector_get"
 
-    vect: Operand = operand_def(VectorType)
-    element: Operand = operand_def(i32)
+    vect = operand_def(VectorType)
+    element = operand_def(i32)
 
-    result: OpResult = result_def(VectorWrappable)
+    result = result_def(VectorWrappableConstr)
 
-    @staticmethod
-    def get(vect: SSAValue | Operation, element: SSAValue | Operation) -> VectorGetOp:
+    def __init__(self, vect: SSAValue | Operation, element: SSAValue | Operation):
         ssa_val = SSAValue.get(vect)
         assert isa(ssa_val.type, VectorType[VectorWrappable])
 
-        return VectorGetOp.build(
+        return super().__init__(
             result_types=[ssa_val.type.wrapped_type], operands=[vect, element]
         )
 
@@ -773,11 +774,10 @@ class NullRequestOp(MPIBaseOp):
 
     name = "mpi.request_null"
 
-    request: Operand = operand_def(RequestType)
+    request = operand_def(RequestType)
 
-    @staticmethod
-    def get(req: SSAValue | Operation):
-        return NullRequestOp.build(operands=[req])
+    def __init__(self, req: SSAValue | Operation):
+        return super().__init__(operands=[req])
 
 
 @irdl_op_definition
@@ -804,15 +804,15 @@ class GatherOp(MPIBaseOp):
 
     name = "mpi.gather"
 
-    sendbuf: Operand = operand_def(llvm.LLVMPointerType)
-    sendcount: Operand = operand_def(i32)
-    sendtype: Operand = operand_def(DataType)
+    sendbuf = operand_def(llvm.LLVMPointerType)
+    sendcount = operand_def(i32)
+    sendtype = operand_def(DataType)
 
-    recvbuf: Operand = operand_def(llvm.LLVMPointerType)
-    recvcount: Operand = operand_def(i32)
-    recvtype: Operand = operand_def(DataType)
+    recvbuf = operand_def(llvm.LLVMPointerType)
+    recvcount = operand_def(i32)
+    recvtype = operand_def(DataType)
 
-    root: Operand = operand_def(i32)
+    root = operand_def(i32)
 
     def __init__(
         self,
@@ -838,22 +838,24 @@ class GatherOp(MPIBaseOp):
 
 
 MPI = Dialect(
+    "mpi",
     [
-        Isend,
-        Irecv,
-        Test,
-        Recv,
-        Send,
-        Reduce,
-        Allreduce,
-        Bcast,
-        Wait,
-        Waitall,
-        GetStatusField,
-        Init,
-        Finalize,
-        CommRank,
-        UnwrapMemrefOp,
+        CommSizeOp,
+        IsendOp,
+        IrecvOp,
+        TestOp,
+        RecvOp,
+        SendOp,
+        ReduceOp,
+        AllreduceOp,
+        BcastOp,
+        WaitOp,
+        WaitallOp,
+        GetStatusFieldOp,
+        InitOp,
+        FinalizeOp,
+        CommRankOp,
+        UnwrapMemRefOp,
         GetDtypeOp,
         AllocateTypeOp,
         VectorGetOp,
