@@ -1,20 +1,23 @@
 from __future__ import annotations
 
-from typing import Annotated, Generic, Iterable, Sequence, TypeVar
+from collections.abc import Iterable, Sequence
+from typing import Generic
+
+from typing_extensions import TypeVar
 
 from xdsl.dialects.builtin import (
-    AnyArrayAttr,
+    I16,
+    I32,
     ArrayAttr,
     IntegerAttr,
     IntegerType,
     StringAttr,
-    i32,
 )
 from xdsl.ir import (
     Attribute,
     Block,
     Dialect,
-    OpResult,
+    Operation,
     ParametrizedAttribute,
     Region,
     SSAValue,
@@ -23,25 +26,22 @@ from xdsl.ir import (
 from xdsl.irdl import (
     AttrSizedOperandSegments,
     IRDLOperation,
-    Operand,
-    OptOperand,
-    OptRegion,
-    ParameterDef,
-    VarOperand,
-    VarOpResult,
-    attr_def,
+    base,
     irdl_attr_definition,
     irdl_op_definition,
+    lazy_traits_def,
     operand_def,
-    opt_attr_def,
     opt_operand_def,
+    opt_prop_def,
     opt_region_def,
+    prop_def,
     region_def,
     result_def,
+    traits_def,
     var_operand_def,
     var_result_def,
 )
-from xdsl.parser import Parser
+from xdsl.parser import AttrParser, Parser
 from xdsl.printer import Printer
 from xdsl.traits import HasParent, IsTerminator, NoTerminator, OptionalSymbolOpInterface
 from xdsl.utils.exceptions import VerifyException
@@ -77,9 +77,36 @@ def parse_operands_with_types(parser: Parser) -> list[SSAValue]:
 
 
 def print_operands_with_types(printer: Printer, operands: Iterable[SSAValue]) -> None:
-    printer.print_list(operands, printer.print)
-    printer.print(" : ")
-    printer.print_list([operand.type for operand in operands], printer.print)
+    printer.print_list(operands, printer.print_ssa_value)
+    printer.print_string(" : ")
+    printer.print_list(operands, lambda o: printer.print_attribute(o.type))
+
+
+def has_binding_use(op: Operation) -> bool:
+    """
+    Returns true if the given operation is used by a "binding" pdl operation.
+    """
+    for result in op.results:
+        for use in result.uses:
+            if not isinstance(use.operation, ResultOp | ResultsOp) or has_binding_use(
+                use.operation
+            ):
+                return True
+    return False
+
+
+def verify_has_binding_use(op: Operation) -> None:
+    """
+    Raise an exception if the operation is in the main matcher body and is
+    not used by a "binding" pdl operation.
+    """
+    if not isinstance(op.parent_op(), PatternOp):
+        return
+    if not has_binding_use(op):
+        raise VerifyException(
+            "expected a bindable user when defined in the matcher body of a "
+            "`pdl.pattern`"
+        )
 
 
 @irdl_attr_definition
@@ -103,6 +130,9 @@ class ValueType(ParametrizedAttribute, TypeAttribute):
 
 
 AnyPDLType = AttributeType | OperationType | TypeType | ValueType
+AnyPDLTypeConstr = (
+    base(AttributeType) | base(OperationType) | base(TypeType) | base(ValueType)
+)
 
 _RangeT = TypeVar(
     "_RangeT",
@@ -112,54 +142,76 @@ _RangeT = TypeVar(
 
 
 @irdl_attr_definition
-class RangeType(Generic[_RangeT], ParametrizedAttribute, TypeAttribute):
+class RangeType(ParametrizedAttribute, TypeAttribute, Generic[_RangeT]):
     name = "pdl.range"
-    element_type: ParameterDef[_RangeT]
+    element_type: _RangeT
 
-    def __init__(self, element_type: _RangeT):
-        super().__init__([element_type])
+    @classmethod
+    def parse_parameters(cls, parser: AttrParser) -> Sequence[Attribute]:
+        parser.parse_punctuation("<")
+        if parser.parse_optional_keyword("attribute") is not None:
+            element_type = AttributeType()
+        elif parser.parse_optional_keyword("operation") is not None:
+            element_type = OperationType()
+        elif parser.parse_optional_keyword("type") is not None:
+            element_type = TypeType()
+        elif parser.parse_optional_keyword("value") is not None:
+            element_type = ValueType()
+        else:
+            parser.raise_error("expected PDL element type for range")
+        parser.parse_punctuation(">")
+        return [element_type]
+
+    def print_parameters(self, printer: Printer) -> None:
+        match self.element_type:
+            case AttributeType():
+                printer.print_string("<attribute>")
+            case OperationType():
+                printer.print_string("<operation>")
+            case TypeType():
+                printer.print_string("<type>")
+            case ValueType():
+                printer.print_string("<value>")
 
 
 @irdl_op_definition
 class ApplyNativeConstraintOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlapply_native_constraint-mlirpdlapplynativeconstraintop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlapply_native_constraint-pdlapplynativeconstraintop).
     """
 
     name = "pdl.apply_native_constraint"
-    constraint_name: StringAttr = attr_def(StringAttr, attr_name="name")
-    args: VarOperand = var_operand_def(AnyPDLType)
+    constraint_name = prop_def(StringAttr, prop_name="name")
+    args = var_operand_def(AnyPDLTypeConstr)
+    res = var_result_def(AnyPDLTypeConstr)
 
-    def __init__(self, name: str | StringAttr, args: Sequence[SSAValue]) -> None:
+    assembly_format = (
+        "$name (`(` $args^ `:` type($args) `)`)? (`:` type($res)^)? attr-dict"
+    )
+
+    def __init__(
+        self,
+        name: str | StringAttr,
+        args: Sequence[SSAValue],
+        result_types: Sequence[Attribute],
+    ) -> None:
         if isinstance(name, str):
             name = StringAttr(name)
-        super().__init__(operands=[args], attributes={"name": name})
-
-    @classmethod
-    def parse(cls, parser: Parser) -> ApplyNativeConstraintOp:
-        name = parser.parse_str_literal()
-        parser.parse_punctuation("(")
-        operands = parse_operands_with_types(parser)
-        parser.parse_punctuation(")")
-        return ApplyNativeConstraintOp(name, operands)
-
-    def print(self, printer: Printer) -> None:
-        printer.print_string_literal(self.constraint_name.data)
-        printer.print("(")
-        print_operands_with_types(printer, self.operands)
-        printer.print(")")
+        super().__init__(
+            result_types=[result_types], operands=[args], properties={"name": name}
+        )
 
 
 @irdl_op_definition
 class ApplyNativeRewriteOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlapply_native_rewrite-mlirpdlapplynativerewriteop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlapply_native_rewrite-pdlapplynativerewriteop).
     """
 
     name = "pdl.apply_native_rewrite"
-    constraint_name: StringAttr = attr_def(StringAttr, attr_name="name")
-    args: VarOperand = var_operand_def(AnyPDLType)
-    res: VarOpResult = var_result_def(AnyPDLType)
+    constraint_name = prop_def(StringAttr, prop_name="name")
+    args = var_operand_def(AnyPDLTypeConstr)
+    res = var_result_def(AnyPDLTypeConstr)
 
     def __init__(
         self,
@@ -172,7 +224,7 @@ class ApplyNativeRewriteOp(IRDLOperation):
         super().__init__(
             result_types=[result_types],
             operands=[args],
-            attributes={"name": name},
+            properties={"name": name},
         )
 
     @classmethod
@@ -189,26 +241,28 @@ class ApplyNativeRewriteOp(IRDLOperation):
         return ApplyNativeRewriteOp(name, operands, result_types)
 
     def print(self, printer: Printer) -> None:
-        printer.print(" ")
+        printer.print_string(" ")
         printer.print_string_literal(self.constraint_name.data)
-        printer.print("(")
-        print_operands_with_types(printer, self.operands)
-        printer.print(")")
+        with printer.in_parens():
+            print_operands_with_types(printer, self.operands)
         if len(self.results) != 0:
-            printer.print(" : ")
-            printer.print_list(self.results, printer.print)
+            printer.print_string(" : ")
+            printer.print_list(self.result_types, printer.print_attribute)
 
 
 @irdl_op_definition
 class AttributeOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlattribute-mlirpdlattributeop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlattribute-pdlattributeop).
+    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlattribute-pdlattributeop
     """
 
     name = "pdl.attribute"
-    value: Attribute | None = opt_attr_def(Attribute)
-    value_type: OptOperand = opt_operand_def(TypeType)
-    output: OpResult = result_def(AttributeType)
+    value = opt_prop_def()
+    value_type = opt_operand_def(TypeType)
+    output = result_def(AttributeType)
+
+    assembly_format = "(`:` $value_type^)? (`=` $value^)? attr-dict-with-keyword"
 
     def verify_(self):
         if self.value is not None and self.value_type is not None:
@@ -216,127 +270,96 @@ class AttributeOp(IRDLOperation):
                 f"{self.name} cannot both specify an expected attribute "
                 "via a constant value and an expected type."
             )
+        if self.value is None and isinstance(self.parent_op(), RewriteOp):
+            raise VerifyException(
+                "expected constant value when specified within a `pdl.rewrite`"
+            )
+        verify_has_binding_use(self)
 
     def __init__(self, value: Attribute | SSAValue | None = None) -> None:
         """
         The given value is either the expected attribute, if given an attribute, or the
         expected attribute type, if given an SSAValue.
         """
-        attributes: dict[str, Attribute] = {}
+        properties: dict[str, Attribute] = {}
         operands: list[SSAValue | None] = [None]
         if isinstance(value, Attribute):
-            attributes["value"] = value
+            properties["value"] = value
         elif isinstance(value, SSAValue):
             operands = [value]
 
         super().__init__(
-            operands=operands, attributes=attributes, result_types=[AttributeType()]
+            operands=operands, properties=properties, result_types=[AttributeType()]
         )
 
-    @classmethod
-    def parse(cls, parser: Parser) -> AttributeOp:
-        value: Operand | Attribute | None = None
-        if parser.parse_optional_punctuation(":"):
-            value = parser.parse_operand()
-        elif parser.parse_optional_punctuation("="):
-            value = parser.parse_attribute()
-
-        return AttributeOp(value)
-
-    def print(self, printer: Printer) -> None:
-        if self.value is not None:
-            printer.print(" = ", self.value)
-        elif self.value_type is not None:
-            printer.print(" : ", self.value_type)
-
 
 @irdl_op_definition
 class EraseOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlerase-mlirpdleraseop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlerase-pdleraseop).
     """
 
     name = "pdl.erase"
-    op_value: Operand = operand_def(OperationType)
+    op_value = operand_def(OperationType)
+
+    assembly_format = "$op_value attr-dict"
 
     def __init__(self, op_value: SSAValue) -> None:
         super().__init__(operands=[op_value])
 
-    @classmethod
-    def parse(cls, parser: Parser) -> EraseOp:
-        op_value = parser.parse_operand()
-        return EraseOp(op_value)
-
-    def print(self, printer: Printer) -> None:
-        printer.print(" ", self.op_value)
-
 
 @irdl_op_definition
 class OperandOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdloperand-mlirpdloperandop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdloperand-pdloperandop).
     """
 
     name = "pdl.operand"
-    value_type: OptOperand = opt_operand_def(TypeType)
-    value: OpResult = result_def(ValueType)
+    value_type = opt_operand_def(TypeType)
+    value = result_def(ValueType)
+
+    assembly_format = "(`:` $value_type^)? attr-dict"
 
     def __init__(self, value_type: SSAValue | None = None) -> None:
         super().__init__(operands=[value_type], result_types=[ValueType()])
 
-    @classmethod
-    def parse(cls, parser: Parser) -> OperandOp:
-        value = None
-        if parser.parse_optional_punctuation(":") is not None:
-            value = parser.parse_operand()
-
-        return OperandOp(value)
-
-    def print(self, printer: Printer) -> None:
-        if self.value_type is not None:
-            printer.print(" : ", self.value_type)
+    def verify_(self):
+        verify_has_binding_use(self)
 
 
 @irdl_op_definition
 class OperandsOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdloperands-mlirpdloperandsop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdloperands-pdloperandsop).
     """
 
     name = "pdl.operands"
-    value_type: OptOperand = opt_operand_def(RangeType[TypeType])
-    value: OpResult = result_def(RangeType[ValueType])
+    value_type = opt_operand_def(RangeType[TypeType])
+    value = result_def(RangeType[ValueType])
+
+    assembly_format = "(`:` $value_type^)? attr-dict"
 
     def __init__(self, value_type: SSAValue | None) -> None:
         super().__init__(operands=[value_type], result_types=[RangeType(ValueType())])
 
-    @classmethod
-    def parse(cls, parser: Parser) -> OperandsOp:
-        value_type = None
-        if parser.parse_optional_punctuation(":") is not None:
-            value_type = parser.parse_operand()
-
-        return OperandsOp(value_type)
-
-    def print(self, printer: Printer) -> None:
-        if self.value_type is not None:
-            printer.print(" : ", self.value_type)
+    def verify_(self):
+        verify_has_binding_use(self)
 
 
 @irdl_op_definition
 class OperationOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdloperation-mlirpdloperationop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdloperation-pdloperationop).
     """
 
     name = "pdl.operation"
-    opName: StringAttr | None = opt_attr_def(StringAttr)
-    attributeValueNames: ArrayAttr[StringAttr] = attr_def(ArrayAttr[StringAttr])
+    opName = opt_prop_def(StringAttr)
+    attributeValueNames = prop_def(ArrayAttr[StringAttr])
 
-    operand_values: VarOperand = var_operand_def(ValueType | RangeType[ValueType])
-    attribute_values: VarOperand = var_operand_def(AttributeType)
-    type_values: VarOperand = var_operand_def(TypeType | RangeType[TypeType])
-    op: OpResult = result_def(OperationType)
+    operand_values = var_operand_def(base(ValueType) | base(RangeType[ValueType]))
+    attribute_values = var_operand_def(AttributeType)
+    type_values = var_operand_def(base(TypeType) | base(RangeType[TypeType]))
+    op = result_def(OperationType)
 
     irdl_options = [AttrSizedOperandSegments()]
 
@@ -365,12 +388,26 @@ class OperationOp(IRDLOperation):
         super().__init__(
             operands=[operand_values, attribute_values, type_values],
             result_types=[OperationType()],
-            attributes={
+            properties={
                 "attributeValueNames": attribute_value_names,
                 "opName": op_name,
             },
         )
 
+    def verify_(self):
+        is_within_rewrite: bool = isinstance(self.parent_op(), RewriteOp)
+        if is_within_rewrite and self.opName is None:
+            raise VerifyException(
+                "must have an operation name when nested within a `pdl.rewrite`"
+            )
+        if len(self.attributeValueNames) != len(self.attribute_values):
+            raise VerifyException(
+                "expected the same number of attribute values and attribute "
+                f"names, got {len(self.attributeValueNames)} names and "
+                f"{len(self.attribute_values)} values"
+            )
+        verify_has_binding_use(self)
+
     @classmethod
     def parse(cls, parser: Parser) -> OperationOp:
         name = parser.parse_optional_str_literal()
@@ -403,44 +440,83 @@ class OperationOp(IRDLOperation):
 
     def print(self, printer: Printer) -> None:
         if self.opName is not None:
-            printer.print(" ", self.opName)
+            printer.print_string(" ")
+            printer.print_attribute(self.opName)
 
         if len(self.operand_values) != 0:
-            printer.print(" (")
-            print_operands_with_types(printer, self.operand_values)
-            printer.print(")")
+            printer.print_string(" ")
+            with printer.in_parens():
+                print_operands_with_types(printer, self.operand_values)
 
         def print_attribute_entry(entry: tuple[StringAttr, SSAValue]):
-            printer.print(entry[0], " = ", entry[1])
+            printer.print_attribute(entry[0])
+            printer.print_string(" = ")
+            printer.print_ssa_value(entry[1])
 
         if len(self.attributeValueNames) != 0:
-            printer.print(" {")
-            printer.print_list(
-                zip(self.attributeValueNames, self.attribute_values),
-                print_attribute_entry,
-            )
-            printer.print("}")
+            printer.print_string(" ")
+            with printer.in_braces():
+                printer.print_list(
+                    zip(self.attributeValueNames, self.attribute_values),
+                    print_attribute_entry,
+                )
 
         if len(self.type_values) != 0:
-            printer.print(" -> (")
-            print_operands_with_types(printer, self.type_values)
-            printer.print(")")
+            printer.print_string(" -> ")
+            with printer.in_parens():
+                print_operands_with_types(printer, self.type_values)
+
+
+def _visit_pdl_ops(op: Operation, visited: set[Operation]):
+    """
+    Visit all pdl.operands, pdl.results, and pdl.operations connected to the
+    given operation in a pdl.pattern.
+    """
+    # We only look at operations within a `pdl.pattern`.
+    if not isinstance(op.parent_op(), PatternOp):
+        return
+
+    if op in visited:
+        return
+
+    visited.add(op)
+
+    # Traverse the operands
+    if isinstance(op, OperationOp):
+        for value in op.operand_values:
+            assert isinstance(owner := value.owner, Operation)
+            _visit_pdl_ops(owner, visited)
+    elif isinstance(op, ResultOp | ResultsOp):
+        assert isinstance(owner := op.parent_.owner, Operation)
+        _visit_pdl_ops(owner, visited)
+
+    # Traverse the users
+    for result in op.results:
+        for user in result.uses:
+            _visit_pdl_ops(user.operation, visited)
+
+
+def _has_user_in_rewrite(op: Operation) -> bool:
+    """Check if an operation has a user in a pdl.rewrite"""
+    for result in op.results:
+        for use in result.uses:
+            if isinstance(use.operation.parent_op(), RewriteOp):
+                return True
+    return False
 
 
 @irdl_op_definition
 class PatternOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlpattern-mlirpdlpatternop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlpattern-pdlpatternop).
     """
 
     name = "pdl.pattern"
-    benefit: IntegerAttr[Annotated[IntegerType, IntegerType(16)]] = attr_def(
-        IntegerAttr[Annotated[IntegerType, IntegerType(16)]]
-    )
-    sym_name: StringAttr | None = opt_attr_def(StringAttr)
-    body: Region = region_def()
+    benefit = prop_def(IntegerAttr[I16])
+    sym_name = opt_prop_def(StringAttr)
+    body = region_def("single_block")
 
-    traits = frozenset([OptionalSymbolOpInterface()])
+    traits = traits_def(OptionalSymbolOpInterface())
 
     def __init__(
         self,
@@ -455,7 +531,7 @@ class PatternOp(IRDLOperation):
         if body is None:
             body = Region(Block())
         super().__init__(
-            attributes={
+            properties={
                 "benefit": benefit,
                 "sym_name": sym_name,
             },
@@ -463,6 +539,39 @@ class PatternOp(IRDLOperation):
             result_types=[],
         )
 
+    def verify_(self):
+        # Check for the correct terminator.
+        if not isinstance(self.body.block.last_op, RewriteOp):
+            raise VerifyException("expected body to terminate with a `pdl.rewrite`")
+
+        # Check that there is at least one `pdl.operation`.
+        if not any(isinstance(op, OperationOp) for op in self.body.block.ops):
+            raise VerifyException(
+                "the pattern must contain at least one `pdl.operation`"
+            )
+
+        # Get the connected component by traversing the graph in the first
+        # PDL operation, operand, or result used in a `pdl.rewrite`. The other
+        # operations will be detected via other means with a better error handling
+        # (expected bindable user).
+        first = True
+        visited: set[Operation] = set()
+        for op in self.body.block.ops:
+            if not isinstance(
+                op, OperandOp | OperandsOp | ResultOp | ResultsOp | OperationOp
+            ):
+                continue
+            if not _has_user_in_rewrite(op):
+                continue
+
+            if first:
+                first = False
+                _visit_pdl_ops(op, visited)
+            if op not in visited:
+                raise VerifyException(
+                    "Operations in a `pdl.pattern` must form a connected component"
+                )
+
     @classmethod
     def parse(cls, parser: Parser) -> PatternOp:
         name = parser.parse_optional_symbol_name()
@@ -476,19 +585,23 @@ class PatternOp(IRDLOperation):
 
     def print(self, printer: Printer) -> None:
         if self.sym_name is not None:
-            printer.print(" @", self.sym_name.data)
-        printer.print(" : benefit(", self.benefit.value.data, ") ", self.body)
+            printer.print_string(" @")
+            printer.print_string(self.sym_name.data)
+        printer.print_string(f" : benefit({self.benefit.value.data}) ")
+        printer.print_region(self.body)
 
 
 @irdl_op_definition
 class RangeOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlrange-mlirpdlrangeop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlrange-pdlrangeop).
     """
 
     name = "pdl.range"
-    arguments: VarOperand = var_operand_def(AnyPDLType | RangeType[AnyPDLType])
-    result: OpResult = result_def(RangeType[AnyPDLType])
+    arguments = var_operand_def(AnyPDLTypeConstr | base(RangeType[AnyPDLType]))
+    result = result_def(RangeType[AnyPDLType])
+
+    traits = lazy_traits_def(lambda: (HasParent(RewriteOp),))
 
     def verify_(self) -> None:
         def get_type_or_elem_type(arg: SSAValue) -> Attribute:
@@ -519,7 +632,7 @@ class RangeOp(IRDLOperation):
 
             if isa(arguments[0].type, RangeType[AnyPDLType]):
                 result_type = RangeType(arguments[0].type.element_type)
-            elif isa(arguments[0].type, AnyPDLType):
+            elif AnyPDLTypeConstr.verifies(arguments[0].type):
                 result_type = RangeType(arguments[0].type)
             else:
                 raise ValueError(
@@ -538,18 +651,19 @@ class RangeOp(IRDLOperation):
 
     def print(self, printer: Printer) -> None:
         if len(self.arguments) == 0:
-            printer.print(" : ", self.result.type)
+            printer.print_string(" : ")
+            printer.print_attribute(self.result.type)
             return
-
+        printer.print_string(" ")
         print_operands_with_types(printer, self.arguments)
 
 
 @irdl_op_definition
 class ReplaceOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlreplace-mlirpdlreplaceop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlreplace-pdlreplaceop).
 
-    pdl.replace` operations are used within `pdl.rewrite` regions to specify
+    `pdl.replace` operations are used within `pdl.rewrite` regions to specify
     that an input operation should be marked as replaced. The semantics of this
     operation correspond with the `replaceOp` method on a `PatternRewriter`. The
     set of replacement values can be either:
@@ -560,12 +674,17 @@ class ReplaceOp(IRDLOperation):
     """
 
     name = "pdl.replace"
-    op_value: Operand = operand_def(OperationType)
-    repl_operation: OptOperand = opt_operand_def(OperationType)
-    repl_values: VarOperand = var_operand_def(ValueType | ArrayAttr[ValueType])
+    op_value = operand_def(OperationType)
+    repl_operation = opt_operand_def(OperationType)
+    repl_values = var_operand_def(base(ValueType) | base(ArrayAttr[ValueType]))
 
     irdl_options = [AttrSizedOperandSegments()]
 
+    assembly_format = (
+        "$op_value `with` ` ` "
+        "(`(` $repl_values^ `:` type($repl_values) `)`)? $repl_operation attr-dict"
+    )
+
     def __init__(
         self,
         op_value: SSAValue,
@@ -596,69 +715,38 @@ class ReplaceOp(IRDLOperation):
                 "`ReplaceOp`, both are set"
             )
 
-    @classmethod
-    def parse(cls, parser: Parser) -> ReplaceOp:
-        root = parser.parse_operand()
-        parser.parse_keyword("with")
-        if (repl_op := parser.parse_optional_operand()) is not None:
-            return ReplaceOp(root, repl_op)
-
-        parser.parse_punctuation("(")
-        repl_values = parse_operands_with_types(parser)
-        parser.parse_punctuation(")")
-        return ReplaceOp(root, repl_values=repl_values)
-
-    def printer(self, printer: Printer) -> None:
-        printer.print(self.op_value, " with ")
-        if self.repl_operation is not None:
-            printer.print(self.repl_operation)
-            return
-        printer.print("(")
-        print_operands_with_types(printer, self.repl_values)
-        printer.print(")")
-
 
 @irdl_op_definition
 class ResultOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlresult-mlirpdlresultop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlresult-pdlresultop).
     """
 
     name = "pdl.result"
-    index: IntegerAttr[Annotated[IntegerType, i32]] = attr_def(
-        IntegerAttr[Annotated[IntegerType, i32]]
-    )
-    parent_: Operand = operand_def(OperationType)
-    val: OpResult = result_def(ValueType)
+    index = prop_def(IntegerAttr[I32])
+    parent_ = operand_def(OperationType)
+    val = result_def(ValueType)
+
+    assembly_format = "$index `of` $parent_ attr-dict"
 
     def __init__(self, index: int | IntegerAttr[IntegerType], parent: SSAValue) -> None:
         if isinstance(index, int):
             index = IntegerAttr(index, 32)
         super().__init__(
-            operands=[parent], attributes={"index": index}, result_types=[ValueType()]
+            operands=[parent], properties={"index": index}, result_types=[ValueType()]
         )
 
-    @classmethod
-    def parse(cls, parser: Parser) -> ResultOp:
-        index = parser.parse_integer()
-        parser.parse_keyword("of")
-        parent = parser.parse_operand()
-        return ResultOp(index, parent)
-
-    def print(self, printer: Printer) -> None:
-        printer.print(" ", self.index.value.data, " of ", self.parent_)
-
 
 @irdl_op_definition
 class ResultsOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlresults-mlirpdlresultsop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlresults-pdlresultsop).
     """
 
     name = "pdl.results"
-    index: IntegerAttr[IntegerType] | None = opt_attr_def(IntegerAttr[IntegerType])
-    parent_: Operand = operand_def(OperationType)
-    val: OpResult = result_def(ValueType | RangeType[ValueType])
+    index = opt_prop_def(IntegerAttr[I32])
+    parent_ = operand_def(OperationType)
+    val = result_def(base(ValueType) | base(RangeType[ValueType]))
 
     def __init__(
         self,
@@ -669,7 +757,7 @@ class ResultsOp(IRDLOperation):
         if isinstance(index, int):
             index = IntegerAttr(index, 32)
         super().__init__(
-            operands=[parent], result_types=[result_type], attributes={"index": index}
+            operands=[parent], result_types=[result_type], properties={"index": index}
         )
 
     @classmethod
@@ -686,31 +774,39 @@ class ResultsOp(IRDLOperation):
 
     def print(self, printer: Printer) -> None:
         if self.index is None:
-            printer.print(" of ", self.parent_)
-            return
-        printer.print(
-            " ", self.index.value.data, " of ", self.parent_, " -> ", self.val.type
-        )
+            printer.print_string(" of ")
+            printer.print_ssa_value(self.parent_)
+        else:
+            printer.print_string(f" {self.index.value.data} of ")
+            printer.print_ssa_value(self.parent_)
+            printer.print_string(" -> ")
+            printer.print_attribute(self.val.type)
 
 
 @irdl_op_definition
 class RewriteOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlrewrite-mlirpdlrewriteop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlrewrite-pdlrewriteop).
     """
 
     name = "pdl.rewrite"
-    root: OptOperand = opt_operand_def(OperationType)
+    root = opt_operand_def(OperationType)
     # name of external rewriter function
-    name_: StringAttr | None = opt_attr_def(StringAttr, attr_name="name")
+    name_ = opt_prop_def(StringAttr, prop_name="name")
     # parameters of external rewriter function
-    external_args: VarOperand = var_operand_def(AnyPDLType)
+    external_args = var_operand_def(AnyPDLTypeConstr)
     # body of inline rewriter function
-    body: OptRegion = opt_region_def()
+    body = opt_region_def()
 
     irdl_options = [AttrSizedOperandSegments()]
 
-    traits = frozenset([HasParent(PatternOp), NoTerminator(), IsTerminator()])
+    traits = traits_def(HasParent(PatternOp), NoTerminator(), IsTerminator())
+
+    assembly_format = (
+        "($root^)? "
+        "(`with` $name^ (`(` $external_args^ `:` type($external_args) `)`)?)?"
+        "($body^)? attr-dict-with-keyword"
+    )
 
     def __init__(
         self,
@@ -737,111 +833,67 @@ class RewriteOp(IRDLOperation):
         elif body is None:
             regions.append([])
 
-        attributes: dict[str, Attribute] = {}
+        properties: dict[str, Attribute] = {}
         if name is not None:
-            attributes["name"] = name
+            properties["name"] = name
 
         super().__init__(
             result_types=[],
             operands=operands,
-            attributes=attributes,
+            properties=properties,
             regions=regions,
         )
 
-    @classmethod
-    def parse(cls, parser: Parser) -> RewriteOp:
-        root = parser.parse_optional_operand()
-
-        if parser.parse_optional_keyword("with") is None:
-            body = parser.parse_region()
-            return RewriteOp(root, body)
-
-        name = parser.parse_str_literal()
-        external_args = []
-        if parser.parse_optional_punctuation("(") is not None:
-            external_args = parse_operands_with_types(parser)
-            parser.parse_punctuation(")")
-
-        return RewriteOp(root, None, name, external_args)
-
-    def print(self, printer: Printer) -> None:
-        if self.root is not None:
-            printer.print(" ", self.root)
-
-        if self.body is not None:
-            printer.print(" ", self.body)
-            return
-
-        printer.print(" with ", self.name_)
-        if len(self.external_args) != 0:
-            printer.print("(")
-            print_operands_with_types(printer, self.external_args)
-            printer.print(")")
-
 
 @irdl_op_definition
 class TypeOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdltype-mlirpdltypeop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdltype-pdltypeop).
     """
 
     name = "pdl.type"
-    constantType: Attribute | None = opt_attr_def(Attribute)
-    result: OpResult = result_def(TypeType)
+    constantType = opt_prop_def(TypeAttribute)
+    result = result_def(TypeType)
+
+    assembly_format = "attr-dict (`:` $constantType^)?"
 
     def __init__(self, constant_type: Attribute | None = None) -> None:
         super().__init__(
-            attributes={"constantType": constant_type}, result_types=[TypeType()]
+            properties={"constantType": constant_type}, result_types=[TypeType()]
         )
 
-    @classmethod
-    def parse(cls, parser: Parser) -> TypeOp:
-        if parser.parse_optional_punctuation(":") is None:
-            return TypeOp()
-        constant_type = parser.parse_attribute()
-        return TypeOp(constant_type)
-
-    def print(self, printer: Printer) -> None:
-        if self.constantType is not None:
-            printer.print(" : ", self.constantType)
+    def verify_(self):
+        verify_has_binding_use(self)
 
 
 @irdl_op_definition
 class TypesOp(IRDLOperation):
     """
-    https://mlir.llvm.org/docs/Dialects/PDLOps/#pdltypes-mlirpdltypesop
+    See external [documentation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdltypes-pdltypesop).
     """
 
     name = "pdl.types"
-    constantTypes: AnyArrayAttr | None = opt_attr_def(AnyArrayAttr)
-    result: OpResult = result_def(RangeType[TypeType])
+    constantTypes = opt_prop_def(ArrayAttr[TypeAttribute])
+    result = result_def(RangeType[TypeType])
+
+    assembly_format = "attr-dict (`:` $constantTypes^)?"
 
     def __init__(self, constant_types: Iterable[Attribute] | None = None) -> None:
         if constant_types is not None:
-            attributes = {"constantTypes": ArrayAttr(constant_types)}
+            properties = {"constantTypes": ArrayAttr(constant_types)}
         else:
-            attributes = {}
+            properties = {}
         super().__init__(
-            attributes=attributes,
+            properties=properties,
             result_types=[RangeType(TypeType())],
         )
 
-    @classmethod
-    def parse(cls, parser: Parser) -> TypesOp:
-        if parser.parse_optional_punctuation(":") is None:
-            return TypesOp()
-        begin_attr_pos = parser.pos
-        constant_types = parser.parse_attribute()
-        if not isa(constant_types, AnyArrayAttr):
-            parser.raise_error("Array attribute expected", begin_attr_pos, parser.pos)
-        return TypesOp(constant_types)
-
-    def print(self, printer: Printer) -> None:
-        if self.constantTypes is not None:
-            printer.print(" : ", self.constantTypes)
+    def verify_(self):
+        verify_has_binding_use(self)
 
 
 PDL = Dialect(
+    "pdl",
     [
         ApplyNativeConstraintOp,
         ApplyNativeRewriteOp,
