@@ -1,23 +1,174 @@
 from __future__ import annotations
 
+import platform
+from collections import Counter
+from collections.abc import Callable, Iterable
 from dataclasses import dataclass, field
 from typing import (
     IO,
     Any,
-    Callable,
-    Generator,
-    Iterable,
+    ClassVar,
+    Literal,
     NamedTuple,
     ParamSpec,
     TypeAlias,
-    TypeVar,
 )
 
-from xdsl.dialects.builtin import ModuleOp
-from xdsl.ir import Operation, OperationInvT, SSAValue
-from xdsl.ir.core import Attribute, Block, Region
-from xdsl.traits import CallableOpInterface, IsTerminator, SymbolOpInterface
+from typing_extensions import TypeVar
+
+from xdsl.dialects.builtin import ModuleOp, SymbolRefAttr
+from xdsl.ir import (
+    Attribute,
+    AttributeInvT,
+    Block,
+    Operation,
+    OperationInvT,
+    Region,
+    SSAValue,
+    TypeAttribute,
+)
+from xdsl.traits import (
+    CallableOpInterface,
+    IsTerminator,
+    SymbolTable,
+)
 from xdsl.utils.exceptions import InterpretationError
+from xdsl.utils.scoped_dict import ScopedDict
+
+PythonValues: TypeAlias = tuple[Any, ...]
+"""
+A tuple of result values corresponding to the results of the operation being
+interpreted.
+"""
+
+
+class ReturnedValues(NamedTuple):
+    """
+    If the terminator exits the region being interpreted, such as for function returns,
+    return these values to yield control back to the parent operation of the block.
+    """
+
+    values: PythonValues
+
+
+class Successor(NamedTuple):
+    """
+    If the terminator jumps to another block within the region being interpreted, such
+    as for jumps, return a block along with the values to pass as arguments.
+    """
+
+    block: Block
+    args: PythonValues
+
+
+_FT = TypeVar("_FT", bound="InterpreterFunctions")
+
+TerminatorValue: TypeAlias = ReturnedValues | Successor
+"""
+A terminator operation either yields control to the parent operation or jumps to a
+successor.
+"""
+
+
+class OpImplResult(NamedTuple):
+    """
+    The result of interpreting an Operation. If and only if the Operation is a terminator,
+    it must set the terminator_value.
+    """
+
+    values: PythonValues
+    terminator_value: TerminatorValue | None
+
+
+NonTerminatorOpImpl: TypeAlias = Callable[
+    [_FT, "Interpreter", OperationInvT, PythonValues], PythonValues
+]
+
+TerminatorOpImpl: TypeAlias = Callable[
+    [_FT, "Interpreter", OperationInvT, PythonValues],
+    tuple[TerminatorValue, PythonValues],
+]
+
+OpImpl: TypeAlias = Callable[
+    [_FT, "Interpreter", OperationInvT, PythonValues], OpImplResult
+]
+
+
+# region : TypeAliases
+
+_AttributeInvT0 = TypeVar("_AttributeInvT0", bound=Attribute)
+_AttributeInvT1 = TypeVar("_AttributeInvT1", bound=Attribute)
+CastImpl: TypeAlias = Callable[
+    [_FT, _AttributeInvT0, _AttributeInvT1, Any],
+    Any,
+]
+AttrImpl: TypeAlias = Callable[
+    [_FT, "Interpreter", Attribute, AttributeInvT],
+    Any,
+]
+
+_ImplDict: TypeAlias = dict[type[Operation], OpImpl["InterpreterFunctions", Operation]]
+_IMPL_DICT: TypeAlias = dict[
+    type[Operation],
+    tuple["InterpreterFunctions", OpImpl["InterpreterFunctions", Operation]],
+]
+
+_CastImplDict: TypeAlias = dict[
+    tuple[type[Attribute], type[Attribute]],
+    CastImpl["InterpreterFunctions", Attribute, Attribute],
+]
+_CAST_IMPL_DICT: TypeAlias = dict[
+    tuple[type[Attribute], type[Attribute]],
+    tuple[
+        "InterpreterFunctions", CastImpl["InterpreterFunctions", Attribute, Attribute]
+    ],
+]
+
+_AttrImplDict: TypeAlias = dict[
+    type[Attribute],
+    AttrImpl["InterpreterFunctions", TypeAttribute],
+]
+_ATTR_IMPL_DICT: TypeAlias = dict[
+    type[Attribute],
+    tuple["InterpreterFunctions", AttrImpl["InterpreterFunctions", Attribute]],
+]
+
+ExtFuncImpl: TypeAlias = Callable[
+    [_FT, "Interpreter", Operation, PythonValues],
+    PythonValues,
+]
+
+_ExtFuncImplDict: TypeAlias = dict[
+    str,
+    ExtFuncImpl["InterpreterFunctions"],
+]
+_EXT_FUNC_IMPL_DICT: TypeAlias = dict[
+    str, tuple["InterpreterFunctions", ExtFuncImpl["InterpreterFunctions"]]
+]
+
+_CallableImplDict: TypeAlias = dict[
+    type[Operation], NonTerminatorOpImpl["InterpreterFunctions", Operation]
+]
+_CALLABLE_IMPL_DICT: TypeAlias = dict[
+    type[Operation],
+    tuple[
+        "InterpreterFunctions", NonTerminatorOpImpl["InterpreterFunctions", Operation]
+    ],
+]
+_IMPL_DATA: TypeAlias = dict[type["InterpreterFunctions"], dict[str, Any]]
+
+# endregion
+
+_IMPL_OP_TYPE_KEY = "__impl_op_type"
+_CAST_IMPL_TYPES_KEY = "__cast_impl_types"
+_ATTR_IMPL_TYPES_KEY = "__attr_impl_types"
+_EXT_FUNC_NAME_KEY = "__external_func_name"
+_CALLABLE_OP_TYPE_KEY = "__callable_op_type"
+_IMPL_DICT_KEY = "__impl_dict"
+_CAST_IMPL_DICT_KEY = "__cast_impl_dict"
+_ATTR_IMPL_DICT_KEY = "__attr_impl_dict"
+_EXT_FUNC_DICT_KEY = "__external_func_dict"
+_CALLABLE_IMPL_DICT_KEY = "__callable_impl_dict"
 
 
 @dataclass
@@ -81,7 +232,7 @@ class InterpreterFunctions:
         cls,
     ) -> Iterable[tuple[type[Operation], OpImpl[InterpreterFunctions, Operation]]]:
         try:
-            impl_dict = getattr(cls, _IMPL_DICT)
+            impl_dict = getattr(cls, _IMPL_DICT_KEY)
             return impl_dict.items()
         except AttributeError as e:
             raise ValueError(f"Use `@register_impls` on class {cls.__name__}") from e
@@ -96,18 +247,46 @@ class InterpreterFunctions:
         ]
     ]:
         try:
-            impl_dict = getattr(cls, _CAST_IMPL_DICT)
+            impl_dict = getattr(cls, _CAST_IMPL_DICT_KEY)
             return impl_dict.items()
         except AttributeError as e:
             raise ValueError(f"Use `@register_impls` on class {cls.__name__}") from e
 
+    @classmethod
+    def _attr_impls(
+        cls,
+    ) -> Iterable[
+        tuple[
+            type[Attribute],
+            AttrImpl[InterpreterFunctions, Attribute],
+        ]
+    ]:
+        try:
+            impl_dict = getattr(cls, _ATTR_IMPL_DICT_KEY)
+            return impl_dict.items()
+        except AttributeError as e:
+            raise ValueError(f"Use `@register_impls` on class {cls.__name__}") from e
 
-_FT = TypeVar("_FT", bound=InterpreterFunctions)
+    @classmethod
+    def _ext_impls(
+        cls,
+    ) -> Iterable[tuple[str, ExtFuncImpl[InterpreterFunctions]]]:
+        try:
+            impl_dict = getattr(cls, _EXT_FUNC_DICT_KEY)
+            return impl_dict.items()
+        except AttributeError as e:
+            raise ValueError(f"Use `@register_impls` on class {cls.__name__}") from e
+
+    @classmethod
+    def _callable_impls(
+        cls,
+    ) -> Iterable[tuple[type[Operation], OpImpl[InterpreterFunctions, Operation]]]:
+        try:
+            impl_dict = getattr(cls, _CALLABLE_IMPL_DICT_KEY)
+            return impl_dict.items()
+        except AttributeError as e:
+            raise ValueError(f"Use `@register_impls` on class {cls.__name__}") from e
 
-_IMPL_OP_TYPE = "__impl_op_type"
-_CAST_IMPL_TYPES = "__cast_impl_types"
-_IMPL_DICT = "__impl_dict"
-_CAST_IMPL_DICT = "__cast_impl_dict"
 
 P = ParamSpec("P")
 
@@ -131,14 +310,14 @@ def impl(
         )
 
     def annot(
-        func: NonTerminatorOpImpl[_FT, OperationInvT]
+        func: NonTerminatorOpImpl[_FT, OperationInvT],
     ) -> OpImpl[_FT, OperationInvT]:
         def impl(
             ft: _FT, interpreter: Interpreter, op: OperationInvT, values: PythonValues
         ) -> OpImplResult:
             return OpImplResult(func(ft, interpreter, op, values), None)
 
-        setattr(impl, _IMPL_OP_TYPE, op_type)
+        setattr(impl, _IMPL_OP_TYPE_KEY, op_type)
         return impl
 
     return annot
@@ -169,7 +348,7 @@ def impl_terminator(
             successor, args = func(ft, interpreter, op, values)
             return OpImplResult(args, successor)
 
-        setattr(impl, _IMPL_OP_TYPE, op_type)
+        setattr(impl, _IMPL_OP_TYPE_KEY, op_type)
         return impl
 
     return annot
@@ -191,14 +370,73 @@ def impl_cast(
     """
 
     def annot(
-        func: CastImpl[_FT, _AttributeInvT0, _AttributeInvT1]
+        func: CastImpl[_FT, _AttributeInvT0, _AttributeInvT1],
     ) -> CastImpl[_FT, _AttributeInvT0, _AttributeInvT1]:
-        setattr(func, _CAST_IMPL_TYPES, (input_type, output_type))
+        setattr(func, _CAST_IMPL_TYPES_KEY, (input_type, output_type))
         return func
 
     return annot
 
 
+AttributeInvNoDefaultT = TypeVar("AttributeInvNoDefaultT", bound=Attribute)
+
+
+def impl_attr(
+    input_type: type[AttributeInvNoDefaultT],
+) -> Callable[
+    [AttrImpl[_FT, AttributeInvNoDefaultT]],
+    AttrImpl[_FT, AttributeInvNoDefaultT],
+]:
+    """
+    Marks the conversion from an attribute to a Python value. The
+    `value_for_attribute` method on `Interpreter` will call into this implementation for
+    matching input and output types.
+
+    See `InterpreterFunctions` for more documentation.
+    """
+
+    def annot(func: AttrImpl[_FT, AttributeInvT]) -> AttrImpl[_FT, AttributeInvT]:
+        setattr(func, _ATTR_IMPL_TYPES_KEY, input_type)
+        return func
+
+    return annot
+
+
+def impl_external(
+    sym_name: str,
+) -> Callable[[ExtFuncImpl[_FT]], ExtFuncImpl[_FT]]:
+    """
+    Marks the Python implementation of an external function.
+    """
+
+    def annot(func: ExtFuncImpl[_FT]) -> ExtFuncImpl[_FT]:
+        setattr(func, _EXT_FUNC_NAME_KEY, sym_name)
+        return func
+
+    return annot
+
+
+def impl_callable(
+    op_type: type[OperationInvT],
+) -> Callable[
+    [NonTerminatorOpImpl[_FT, OperationInvT]], NonTerminatorOpImpl[_FT, OperationInvT]
+]:
+    """
+    Marks the Python implementation of a callable operation.
+    """
+
+    if not op_type.has_trait(CallableOpInterface):
+        raise ValueError("Operations that are not callable must use `impl` annotation")
+
+    def annot(
+        impl: NonTerminatorOpImpl[_FT, OperationInvT],
+    ) -> NonTerminatorOpImpl[_FT, OperationInvT]:
+        setattr(impl, _CALLABLE_OP_TYPE_KEY, op_type)
+        return impl
+
+    return annot
+
+
 def register_impls(ft: type[_FT]) -> type[_FT]:
     """
     Enumerates the methods on a given class, and registers the ones marked with
@@ -210,26 +448,50 @@ def register_impls(ft: type[_FT]) -> type[_FT]:
 
     impl_dict: _ImplDict = {}
     cast_impl_dict: _CastImplDict = {}
+    external_func_dict: _ExtFuncImplDict = {}
+    attr_impl_dict: _AttrImplDict = {}
+    callable_impl_dict: _CallableImplDict = {}
 
     for cls in ft.mro():
         # Iterate from subclass through superclasses
         # Assign definitions, unless they've been redefined in a subclass
         for val in cls.__dict__.values():
-            if _IMPL_OP_TYPE in val.__dir__():
+            if _IMPL_OP_TYPE_KEY in val.__dir__():
                 # This is an annotated operation implementation
-                op_type = getattr(val, _IMPL_OP_TYPE)
+                op_type = getattr(val, _IMPL_OP_TYPE_KEY)
                 if op_type not in impl_dict:
                     # subclass overrides superclass definition
                     impl_dict[op_type] = val
-            elif _CAST_IMPL_TYPES in val.__dir__():
+            elif _CAST_IMPL_TYPES_KEY in val.__dir__():
                 # This is an annotated cast implementation
-                types = getattr(val, _CAST_IMPL_TYPES)
+                types = getattr(val, _CAST_IMPL_TYPES_KEY)
                 if types not in cast_impl_dict:
                     # subclass overrides superclass definition
                     cast_impl_dict[types] = val
+            elif _EXT_FUNC_NAME_KEY in val.__dir__():
+                # This is an annotated external function
+                sym_name = getattr(val, _EXT_FUNC_NAME_KEY)
+                assert isinstance(sym_name, str)
+                if sym_name not in external_func_dict:
+                    # subclass overrides superclass definition
+                    external_func_dict[sym_name] = val
+            elif _ATTR_IMPL_TYPES_KEY in val.__dir__():
+                # This is an attribute value implementation
+                types = getattr(val, _ATTR_IMPL_TYPES_KEY)
+                if types not in attr_impl_dict:
+                    # subclass overrides superclass definition
+                    attr_impl_dict[types] = val
+            elif _CALLABLE_OP_TYPE_KEY in val.__dir__():
+                op_type = getattr(val, _CALLABLE_OP_TYPE_KEY)
+                if op_type not in callable_impl_dict:
+                    # subclass overrides superclass definition
+                    callable_impl_dict[op_type] = val
 
-    setattr(ft, _IMPL_DICT, impl_dict)
-    setattr(ft, _CAST_IMPL_DICT, cast_impl_dict)
+    setattr(ft, _IMPL_DICT_KEY, impl_dict)
+    setattr(ft, _CAST_IMPL_DICT_KEY, cast_impl_dict)
+    setattr(ft, _ATTR_IMPL_DICT_KEY, attr_impl_dict)
+    setattr(ft, _EXT_FUNC_DICT_KEY, external_func_dict)
+    setattr(ft, _CALLABLE_IMPL_DICT_KEY, callable_impl_dict)
 
     return ft
 
@@ -242,16 +504,15 @@ class _InterpreterFunctionImpls:
     so we keep a `(Functions, OpImpl)` tuple for every Operation type.
     """
 
-    _impl_dict: dict[
-        type[Operation],
-        tuple[InterpreterFunctions, OpImpl[InterpreterFunctions, Operation]],
-    ] = field(default_factory=dict)
-    _cast_impl_dict: dict[
-        tuple[type[Attribute], type[Attribute]],
-        tuple[
-            InterpreterFunctions, CastImpl[InterpreterFunctions, Attribute, Attribute]
-        ],
-    ] = field(default_factory=dict)
+    _impl_dict: _IMPL_DICT = field(default_factory=_IMPL_DICT)
+    _cast_impl_dict: _CAST_IMPL_DICT = field(default_factory=_CAST_IMPL_DICT)
+    _attr_impl_dict: _ATTR_IMPL_DICT = field(default_factory=_ATTR_IMPL_DICT)
+    _external_funcs_dict: _EXT_FUNC_IMPL_DICT = field(
+        default_factory=_EXT_FUNC_IMPL_DICT
+    )
+    _callable_impl_dict: _CALLABLE_IMPL_DICT = field(
+        default_factory=_CALLABLE_IMPL_DICT
+    )
 
     def register_from(self, ft: InterpreterFunctions, /, override: bool):
         impls = ft._impls()  # pyright: ignore[reportPrivateUsage]
@@ -265,14 +526,44 @@ class _InterpreterFunctionImpls:
             self._impl_dict[op_type] = (ft, impl)
 
         cast_impls = ft._cast_impls()  # pyright: ignore[reportPrivateUsage]
-        for types, impl in cast_impls:
+        for types, cast_impl in cast_impls:
             if types in self._cast_impl_dict and not override:
                 raise ValueError(
                     "Attempting to register implementation for cast with types "
                     f"{types}, but types already registered"
                 )
 
-            self._cast_impl_dict[types] = (ft, impl)
+            self._cast_impl_dict[types] = (ft, cast_impl)
+
+        cast_impls = ft._attr_impls()  # pyright: ignore[reportPrivateUsage]
+        for types, cast_impl in cast_impls:
+            if types in self._attr_impl_dict and not override:
+                raise ValueError(
+                    "Attempting to register implementation for cast with types "
+                    f"{types}, but types already registered"
+                )
+
+            self._attr_impl_dict[types] = (ft, cast_impl)
+
+        ext_impls = ft._ext_impls()  # pyright: ignore[reportPrivateUsage]
+        for sym_name, ext_impl in ext_impls:
+            if sym_name in self._external_funcs_dict and not override:
+                raise ValueError(
+                    "Attempting to register external function with name "
+                    f"{sym_name}, but it's already registered"
+                )
+
+            self._external_funcs_dict[sym_name] = (ft, ext_impl)
+
+        callable_impls = ft._callable_impls()  # pyright: ignore[reportPrivateUsage]
+        for op_type, impl in callable_impls:
+            if op_type in self._callable_impl_dict and not override:
+                raise ValueError(
+                    "Attempting to register implementation for op of type "
+                    f"{op_type}, but type already registered"
+                )
+
+            self._callable_impl_dict[op_type] = (ft, impl)
 
     def run(
         self, interpreter: Interpreter, op: Operation, args: tuple[Any, ...]
@@ -282,7 +573,10 @@ class _InterpreterFunctionImpls:
                 f"Could not find interpretation function for op {op.name}"
             )
         ft, impl = self._impl_dict[type(op)]
-        return impl(ft, interpreter, op, args)
+        try:
+            return impl(ft, interpreter, op, args)
+        except Exception as e:
+            op.emit_error("Error while interpreting op", e)
 
     def cast(
         self,
@@ -298,52 +592,45 @@ class _InterpreterFunctionImpls:
         ft, impl = self._cast_impl_dict[types]
         return impl(ft, input_type, output_type, value)
 
-
-@dataclass
-class InterpreterContext:
-    """
-    Class holding the Python values associated with SSAValues during an
-    interpretation context. An environment is a stack of scopes, values are
-    assigned to the current scope, but can be fetched from a parent scope.
-    """
-
-    name: str = field(default="unknown")
-    parent: InterpreterContext | None = None
-    env: dict[SSAValue, Any] = field(default_factory=dict)
-
-    def __getitem__(self, key: SSAValue) -> Any:
-        """
-        Fetch key from environment. Attempts to first fetch from current scope,
-        then from parent scopes. Raises Interpretation error if not found.
-        """
-        if key in self.env:
-            return self.env[key]
-        if self.parent is not None:
-            return self.parent[key]
-        raise InterpretationError(f"Could not find value for {key} in {self}")
-
-    def __setitem__(self, key: SSAValue, value: Any):
-        """
-        Assign key to current scope. Raises InterpretationError if key already
-        assigned to.
-        """
-        if key in self.env:
+    def attr_value(
+        self, interpreter: Interpreter, attr: Attribute, type_attr: Attribute
+    ) -> Any:
+        attr_type = type(type_attr)
+        if attr_type not in self._attr_impl_dict:
             raise InterpretationError(
-                f"Attempting to register SSAValue {value} for name {key}"
-                f", but value with that name already exists in {self}"
+                f"Could not find Python value implementation for types {attr_type}"
             )
-        self.env[key] = value
+        ft, impl = self._attr_impl_dict[attr_type]
+        return impl(ft, interpreter, attr, type_attr)
 
-    def stack(self) -> Generator[InterpreterContext, None, None]:
-        """
-        Iterates through scopes starting with the root scope.
-        """
-        if self.parent is not None:
-            yield from self.parent.stack()
-        yield self
+    def call_external(
+        self, interpreter: Interpreter, sym_name: str, op: Operation, args: PythonValues
+    ) -> PythonValues:
+        if sym_name not in self._external_funcs_dict:
+            raise InterpretationError(
+                f"Could not find external function implementation named {sym_name}"
+            )
 
-    def __format__(self, __format_spec: str) -> str:
-        return "/".join(c.name for c in self.stack())
+        ft, ext_func = self._external_funcs_dict[sym_name]
+
+        return ext_func(ft, interpreter, op, args)
+
+    def call(
+        self, interpreter: Interpreter, op: Operation, args: PythonValues
+    ) -> PythonValues:
+        ft, ext_func = self._callable_impl_dict[type(op)]
+
+        return ext_func(ft, interpreter, op, args)
+
+
+def _get_system_bitwidth() -> Literal[32, 64] | None:
+    match platform.architecture()[0]:
+        case "64bit":
+            return 64
+        case "32bit":
+            return 32
+        case _:
+            return None
 
 
 @dataclass
@@ -356,25 +643,41 @@ class Interpreter:
     the `register_functions` method.
     """
 
-    module: ModuleOp
-    _impls: _InterpreterFunctionImpls = field(default_factory=_InterpreterFunctionImpls)
-    _ctx: InterpreterContext = field(
-        default_factory=lambda: InterpreterContext(name="root")
+    class Listener:
+        """
+        Base class for observing the operations that are interpreted during a run.
+        """
+
+        def will_interpret_op(self, op: Operation, args: PythonValues) -> None: ...
+
+        def did_interpret_op(self, op: Operation, results: PythonValues) -> None: ...
+
+    SYSTEM_BITWIDTH: ClassVar[Literal[32, 64] | None] = _get_system_bitwidth()
+    DEFAULT_BITWIDTH: ClassVar[Literal[32, 64]] = (
+        32 if SYSTEM_BITWIDTH is None else SYSTEM_BITWIDTH
     )
+
+    module: ModuleOp
+    index_bitwidth: Literal[32, 64] = field(default=DEFAULT_BITWIDTH)
+    """
+    Number of bits in the binary representation of the index
+    """
+    _impls: _InterpreterFunctionImpls = field(default_factory=_InterpreterFunctionImpls)
+    _ctx: ScopedDict[SSAValue, Any] = field(
+        default_factory=lambda: ScopedDict[SSAValue, Any](name="root")
+    )
+    """
+    Object holding the Python values associated with SSAValues during an
+    interpretation context. An environment is a stack of scopes, values are
+    assigned to the current scope, but can be fetched from a parent scope.
+    """
     file: IO[str] | None = field(default=None)
     _symbol_table: dict[str, Operation] | None = None
-
-    @property
-    def symbol_table(self) -> dict[str, Operation]:
-        if self._symbol_table is None:
-            self._symbol_table = {}
-
-            for op in self.module.walk():
-                if (symbol_interface := op.get_trait(SymbolOpInterface)) is not None:
-                    symbol = symbol_interface.get_sym_attr_name(op)
-                    if symbol:
-                        self._symbol_table[symbol.data] = op
-        return self._symbol_table
+    _impl_data: _IMPL_DATA = field(default_factory=_IMPL_DATA)
+    """
+    Runtime data associated with an interpreter functions implementation.
+    """
+    listeners: tuple[Listener, ...] = field(default=())
 
     def get_values(self, values: Iterable[SSAValue]) -> tuple[Any, ...]:
         """
@@ -391,11 +694,11 @@ class Interpreter:
         for ssa_value, result_value in pairs:
             self._ctx[ssa_value] = result_value
 
-    def push_scope(self, name: str = "unknown") -> None:
+    def push_scope(self, name: str | None = None) -> None:
         """
         Create new scope in current environment, with optional custom `name`.
         """
-        self._ctx = InterpreterContext(name, self._ctx)
+        self._ctx = ScopedDict(name=name, parent=self._ctx)
 
     def pop_scope(self) -> None:
         """
@@ -419,40 +722,52 @@ class Interpreter:
         """
         self._impls.register_from(impls, override=override)
 
-    def run_op(self, op: Operation | str, inputs: PythonValues) -> PythonValues:
-        """
-        Calls the implementation for the given operation.
-        """
-        if isinstance(op, str):
-            op = self.get_op_for_symbol(op)
-
+    def _run_op(self, op: Operation, inputs: PythonValues) -> OpImplResult:
+        if (operands_count := len(op.operands)) != (inputs_count := len(inputs)):
+            raise InterpretationError(
+                f"Number of operands ({operands_count}) doesn't match the number of inputs ({inputs_count})."
+            )
+        for listener in self.listeners:
+            listener.will_interpret_op(op, inputs)
         result = self._impls.run(self, op, inputs)
-        return result.values
+        if (results_count := len(op.results)) != (
+            actual_result_count := len(result.values)
+        ):
+            raise InterpretationError(
+                f"Number of operation results ({results_count}) doesn't match the "
+                f"number of implementation results ({actual_result_count})."
+            )
+        for listener in self.listeners:
+            listener.did_interpret_op(op, result.values)
+        return result
 
-    def call_op(self, op: Operation | str, inputs: PythonValues) -> PythonValues:
+    def run_op(
+        self, op: Operation | str | SymbolRefAttr, inputs: PythonValues = ()
+    ) -> PythonValues:
         """
         Calls the implementation for the given operation.
         """
-        if isinstance(op, str):
-            name = op
+        if not isinstance(op, Operation):
             op = self.get_op_for_symbol(op)
-        else:
-            name = "unknown"
 
-        interface = op.get_trait(CallableOpInterface)
+        return self._run_op(op, inputs).values
 
-        self.interpreter_assert(
-            interface is not None,
-            f"Operation {op.name} does not have trait CallableOpInterface",
-        )
-        assert interface is not None
-
-        body = interface.get_callable_region(op)
-
-        results = self.run_ssacfg_region(body, inputs, name)
-        assert results is not None
+    def call_op(
+        self, op: Operation | str | SymbolRefAttr, inputs: PythonValues = ()
+    ) -> PythonValues:
+        """
+        Calls the implementation for the given operation.
+        """
+        if not isinstance(op, Operation):
+            op = self.get_op_for_symbol(op)
+        results = self._impls.call(self, op, inputs)
         return results
 
+    def call_external(
+        self, sym_name: str, op: Operation, inputs: PythonValues = ()
+    ) -> PythonValues:
+        return self._impls.call_external(self, sym_name, op, inputs)
+
     def run_ssacfg_region(
         self, region: Region, args: PythonValues, name: str = "unknown"
     ) -> PythonValues:
@@ -465,12 +780,11 @@ class Interpreter:
         if not region.blocks:
             return results
 
-        scope_count = 0
-        block = region.blocks[0]
+        initial_scope = self._ctx
+        block = region.blocks.first
 
         while block is not None:
             self.push_scope(name)
-            scope_count += 1
             self.set_values(zip(block.args, args))
 
             op: Operation | None = block.first_op
@@ -478,10 +792,10 @@ class Interpreter:
 
             while op is not None:
                 inputs = self.get_values(op.operands)
-                result = self._impls.run(self, op, inputs)
+                result = self._run_op(op, inputs)
                 self.interpreter_assert(
                     len(op.results) == len(result.values),
-                    f"Incorrect number of results for op {op.name}",
+                    f"Incorrect number of results for op {op.name}, expected {len(op.results)} but got {len(result.values)}",
                 )
                 self.set_values(zip(op.results, result.values))
 
@@ -499,9 +813,7 @@ class Interpreter:
                 # Set up next iteration
                 op = op.next_op
 
-        # Pop as many scopes as we entered blocks
-        for _ in range(scope_count):
-            self.pop_scope()
+        self._ctx = initial_scope
         return results
 
     def cast_value(self, o: Attribute, r: Attribute, value: Any) -> Any:
@@ -514,11 +826,52 @@ class Interpreter:
 
         return self._impls.cast(o, r, value)
 
-    def get_op_for_symbol(self, symbol: str) -> Operation:
-        if symbol in self.symbol_table:
-            return self.symbol_table[symbol]
+    def value_for_attribute(self, attr: Attribute, type_attr: Attribute) -> Any:
+        return self._impls.attr_value(self, attr, type_attr)
+
+    def get_op_for_symbol(self, symbol: str | SymbolRefAttr) -> Operation:
+        op = SymbolTable.lookup_symbol(self.module, symbol)
+        if op is not None:
+            return op
+        raise InterpretationError(f"Could not find symbol {symbol}")
+
+    def get_data(
+        self,
+        functions: type[InterpreterFunctions],
+        key: str,
+        factory: Callable[[], Any],
+    ) -> Any:
+        """
+        Get data associated with a specific interpreter functions class, with a given key.
+        If the data is missing, the `factory` argument will be executed and stored on the
+        interpreter.
+        """
+        if functions not in self._impl_data:
+            functions_data: dict[str, Any] = {}
+            self._impl_data[functions] = functions_data
         else:
-            raise InterpretationError(f'Could not find symbol "{symbol}"')
+            functions_data = self._impl_data[functions]
+
+        if key not in functions_data:
+            data = factory()
+            functions_data[key] = data
+        else:
+            data = functions_data[key]
+
+        return data
+
+    def set_data(
+        self,
+        functions: type[InterpreterFunctions],
+        key: str,
+        value: Any,
+    ):
+        if functions not in self._impl_data:
+            functions_data: dict[str, Any] = {}
+            self._impl_data[functions] = functions_data
+        else:
+            functions_data = self._impl_data[functions]
+        functions_data[key] = value
 
     def print(self, *args: Any, **kwargs: Any):
         """Print to current file."""
@@ -527,57 +880,27 @@ class Interpreter:
     def interpreter_assert(self, condition: bool, message: str | None = None):
         """Raise InterpretationError if condition is not satisfied."""
         if not condition:
-            raise InterpretationError(f"AssertionError: ({self._ctx})({message})")
+            self.raise_error(message)
+
+    def scope_names(self):
+        ctx = self._ctx
+
+        while ctx is not None:
+            yield ctx.name or "unknown"
+            ctx = ctx.parent
+
+    def raise_error(self, message: str | None = None):
+        scope_description = "/".join(self.scope_names())
+        raise InterpretationError(f"AssertionError: ({scope_description})({message})")
 
 
-PythonValues: TypeAlias = tuple[Any, ...]
-
-
-class ReturnedValues(NamedTuple):
-    values: PythonValues
-
-
-class Successor(NamedTuple):
-    block: Block
-    args: PythonValues
-
-
-TerminatorValue: TypeAlias = ReturnedValues | Successor
-
-
-class OpImplResult(NamedTuple):
+@dataclass
+class OpCounter(Interpreter.Listener):
     """
-    The result of interpreting an Operation. If and only if the Operation is a terminator,
-    it must set the terminator_value.
+    Counts the number of times that an op has been run by the interpreter.
     """
 
-    values: PythonValues
-    terminator_value: TerminatorValue | None
+    ops: Counter[str] = field(default_factory=Counter[str])
 
-
-NonTerminatorOpImpl: TypeAlias = Callable[
-    [_FT, Interpreter, OperationInvT, PythonValues], PythonValues
-]
-
-TerminatorOpImpl: TypeAlias = Callable[
-    [_FT, Interpreter, OperationInvT, PythonValues],
-    tuple[TerminatorValue, PythonValues],
-]
-
-OpImpl: TypeAlias = Callable[
-    [_FT, Interpreter, OperationInvT, PythonValues], OpImplResult
-]
-
-_AttributeInvT0 = TypeVar("_AttributeInvT0", bound=Attribute)
-_AttributeInvT1 = TypeVar("_AttributeInvT1", bound=Attribute)
-CastImpl: TypeAlias = Callable[
-    [_FT, _AttributeInvT0, _AttributeInvT1, Any],
-    Any,
-]
-
-_ImplDict: TypeAlias = dict[type[Operation], OpImpl[InterpreterFunctions, Operation]]
-
-_CastImplDict: TypeAlias = dict[
-    tuple[type[Attribute], type[Attribute]],
-    CastImpl[InterpreterFunctions, Attribute, Attribute],
-]
+    def will_interpret_op(self, op: Operation, args: PythonValues) -> None:
+        self.ops[op.name] += 1
