@@ -3,27 +3,30 @@ from __future__ import annotations
 import itertools
 import re
 from collections import defaultdict
+from collections.abc import Iterable, Sequence
 from dataclasses import dataclass
-from typing import Any, Iterable, Sequence, cast
+from typing import Literal, overload
 
-from xdsl.dialects.builtin import DictionaryAttr, ModuleOp, UnregisteredAttr
+from xdsl.context import Context
+from xdsl.dialect_interfaces import OpAsmDialectInterface
+from xdsl.dialects.builtin import DictionaryAttr, ModuleOp
 from xdsl.ir import (
     Attribute,
     Block,
-    Data,
-    MLContext,
     Operation,
-    ParametrizedAttribute,
     Region,
     SSAValue,
 )
-from xdsl.parser.attribute_parser import AttrParser
-from xdsl.parser.base_parser import ParserState, Position
+from xdsl.irdl import IRDLOperation
 from xdsl.utils.exceptions import MultipleSpansParseError
-from xdsl.utils.lexer import Input, Lexer, Span, Token
+from xdsl.utils.lexer import Input, Span
+from xdsl.utils.mlir_lexer import MLIRLexer, MLIRTokenKind
+
+from .attribute_parser import AttrParser  # noqa: TID251
+from .generic_parser import ParserState, Position  # noqa: TID251
 
 
-@dataclass
+@dataclass(eq=False)
 class ForwardDeclaredValue(SSAValue):
     """
     An SSA value that is used before it is defined.
@@ -32,13 +35,7 @@ class ForwardDeclaredValue(SSAValue):
 
     @property
     def owner(self) -> Operation | Block:
-        assert False, "Forward declared values do not have an owner"
-
-    def __eq__(self, other: object) -> bool:
-        return self is other
-
-    def __hash__(self) -> int:
-        return id(self)
+        raise ValueError("Forward declared values do not have an owner")
 
 
 @dataclass
@@ -82,7 +79,7 @@ class Parser(AttrParser):
     must_ type parsers are preferred because they are explicit about their failure modes.
     """
 
-    ssa_values: dict[str, tuple[SSAValue]]
+    ssa_values: dict[str, tuple[SSAValue, ...]]
     blocks: dict[str, tuple[Block, Span | None]]
     forward_block_references: dict[str, list[Span]]
     """
@@ -97,14 +94,14 @@ class Parser(AttrParser):
 
     def __init__(
         self,
-        ctx: MLContext,
+        ctx: Context,
         input: str,
         name: str = "<unknown>",
     ) -> None:
-        super().__init__(ParserState(Lexer(Input(input, name))), ctx)
+        super().__init__(ParserState(MLIRLexer(Input(input, name))), ctx)
         self.ssa_values = dict()
         self.blocks = dict()
-        self.forward_block_references = dict()
+        self.forward_block_references = defaultdict(list)
         self.forward_ssa_references = dict()
 
     def parse_module(self, allow_implicit_module: bool = True) -> ModuleOp:
@@ -124,10 +121,20 @@ class Parser(AttrParser):
         else:
             parsed_ops: list[Operation] = []
 
-            while self._current_token.kind != Token.Kind.EOF:
-                if (parsed_op := self.parse_optional_operation()) is None:
-                    self.raise_error("Could not parse entire input!")
-                parsed_ops.append(parsed_op)
+            while self._current_token.kind != MLIRTokenKind.EOF:
+                if self._current_token.kind in (
+                    MLIRTokenKind.HASH_IDENT,
+                    MLIRTokenKind.EXCLAMATION_IDENT,
+                ):
+                    self._parse_alias_def()
+                    continue
+                if self._current_token.kind in (MLIRTokenKind.FILE_METADATA_BEGIN,):
+                    self._parse_file_metadata_dictionary()
+                    continue
+                if (parsed_op := self.parse_optional_operation()) is not None:
+                    parsed_ops.append(parsed_op)
+                    continue
+                self.raise_error("Could not parse entire input!")
 
             if len(parsed_ops) == 0:
                 self.raise_error("Could not parse entire input!")
@@ -142,23 +149,45 @@ class Parser(AttrParser):
             value_names = ", ".join(
                 "%" + name for name in self.forward_ssa_references.keys()
             )
-            if len(self.forward_block_references.keys()) > 1:
-                self.raise_error(f"values {value_names} were used but not defined")
-            else:
-                self.raise_error(f"value {value_names} was used but not defined")
+            self.raise_error(f"values used but not defined: [{value_names}]")
 
         return module_op
 
+    def _parse_alias_def(self):
+        """
+        Parse an attribute or type alias definition with format:
+            alias-def           ::= type-alias-def | attribute-alias-def
+            type-alias-def      ::= `!` bare-id `=` type
+            attribute-alias-def ::= `#` `alias` bare-id `=` attribute
+        """
+        if (
+            token := self._parse_optional_token_in(
+                [MLIRTokenKind.EXCLAMATION_IDENT, MLIRTokenKind.HASH_IDENT]
+            )
+        ) is None:
+            self.raise_error("expected attribute name")
+
+        type_or_attr_name = token.text
+        if type_or_attr_name in self.attribute_aliases:
+            self.raise_error(f"re-declaration of alias '{type_or_attr_name}'")
+
+        self.parse_punctuation("=", "after attribute alias name")
+        value = self.parse_attribute()
+        self.attribute_aliases[type_or_attr_name] = value
+
     def _get_block_from_name(self, block_name: Span) -> Block:
         """
-        This function takes a span containing a block id (like `^42`) and returns a block.
+        This function takes a span containing a block id (like `^bb42`) and returns a block.
 
         If the block definition was not seen yet, we create a forward declaration.
         """
         name = block_name.text[1:]
         if name not in self.blocks:
             self.forward_block_references[name].append(block_name)
-            self.blocks[name] = (Block(), None)
+            block = Block()
+            if Block.is_valid_name(name) and not Block.is_default_block_name(name):
+                block.name_hint = name  # setter verifies validity
+            self.blocks[name] = (block, None)
         return self.blocks[name][0]
 
     def _parse_optional_block_arg_list(self, block: Block):
@@ -168,17 +197,17 @@ class Parser(AttrParser):
             value-id-and-type-list ::= value-id-and-type (`,` ssa-id-and-type)*
             block-arg-list ::= `(` value-id-and-type-list? `)`
         """
-        if self._current_token.kind != Token.Kind.L_PAREN:
+        if self._current_token.kind != MLIRTokenKind.L_PAREN:
             return None
 
         def parse_argument() -> None:
             """Parse a single block argument with its type."""
             arg_name = self._parse_token(
-                Token.Kind.PERCENT_IDENT, "block argument expected"
+                MLIRTokenKind.PERCENT_IDENT, "block argument expected"
             ).span
             self.parse_punctuation(":")
             arg_type = self.parse_attribute()
-            self._parse_optional_location()
+            self.parse_optional_location()
 
             # Insert the block argument in the block, and register it in the parser
             block_arg = block.insert_arg(arg_type, len(block.args))
@@ -203,7 +232,9 @@ class Parser(AttrParser):
           block-id       ::= caret-id
           block-arg-list ::= `(` ssa-id-and-type-list? `)`
         """
-        name_token = self._parse_token(Token.Kind.CARET_IDENT, " in block definition")
+        name_token = self._parse_token(
+            MLIRTokenKind.CARET_IDENT, " in block definition"
+        )
 
         name = name_token.text[1:]
         if name not in self.blocks:
@@ -220,6 +251,11 @@ class Parser(AttrParser):
                 )
             self.forward_block_references.pop(name)
 
+        # Don't set name_hint for blocks that match the default pattern
+        if not Block.is_default_block_name(name):
+            block.name_hint = name  # setter verifies validity
+        # If it matches pattern "bb" followed by digits, leave name_hint as None
+
         self._parse_optional_block_arg_list(block)
         self.parse_punctuation(":")
         self._parse_block_body(block)
@@ -232,12 +268,12 @@ class Parser(AttrParser):
         Parse an operand with format `%<value-id>(#<int-literal>)?`, if present.
         The operand may be forward declared.
         """
-        name_token = self._parse_optional_token(Token.Kind.PERCENT_IDENT)
+        name_token = self._parse_optional_token(MLIRTokenKind.PERCENT_IDENT)
         if name_token is None:
             return None
 
         index = 0
-        index_token = self._parse_optional_token(Token.Kind.HASH_IDENT)
+        index_token = self._parse_optional_token(MLIRTokenKind.HASH_IDENT)
         if index_token is not None:
             if re.fullmatch(self._decimal_integer_regex, index_token.text[1:]) is None:
                 self.raise_error(
@@ -329,106 +365,6 @@ class Parser(AttrParser):
         """Parse an operand with format `%<value-id>`."""
         return self.expect(self.parse_optional_operand, msg)
 
-    def parse_type(self) -> Attribute:
-        """
-        Parse an xDSL type.
-        An xDSL type is either a builtin type, which can have various format,
-        or a dialect type, with the following format:
-            dialect-type  ::= `!` type-name (`<` dialect-type-contents+ `>`)?
-            type-name     ::= bare-id
-            dialect-type-contents ::= `<` dialect-attribute-contents+ `>`
-                            | `(` dialect-attribute-contents+ `)`
-                            | `[` dialect-attribute-contents+ `]`
-                            | `{` dialect-attribute-contents+ `}`
-                            | [^[]<>(){}\0]+
-        """
-        return self.expect(self.parse_optional_type, "type expected")
-
-    def parse_optional_type(self) -> Attribute | None:
-        """
-        Parse an xDSL type, if present.
-        An xDSL type is either a builtin type, which can have various format,
-        or a dialect type, with the following format:
-            dialect-type  ::= `!` type-name (`<` dialect-type-contents+ `>`)?
-            type-name     ::= bare-id
-            dialect-type-contents ::= `<` dialect-attribute-contents+ `>`
-                            | `(` dialect-attribute-contents+ `)`
-                            | `[` dialect-attribute-contents+ `]`
-                            | `{` dialect-attribute-contents+ `}`
-                            | [^[]<>(){}\0]+
-        """
-        if (
-            token := self._parse_optional_token(Token.Kind.EXCLAMATION_IDENT)
-        ) is not None:
-            return self._parse_dialect_type_or_attribute_inner(token.text[1:], True)
-        return self._parse_optional_builtin_type()
-
-    def parse_attribute(self) -> Attribute:
-        """
-        Parse an xDSL attribute.
-        An attribute is either a builtin attribute, which can have various format,
-        or a dialect attribute, with the following format:
-            dialect-attr  ::= `!` attr-name (`<` dialect-attr-contents+ `>`)?
-            attr-name     ::= bare-id
-            dialect-attr-contents ::= `<` dialect-attribute-contents+ `>`
-                            | `(` dialect-attribute-contents+ `)`
-                            | `[` dialect-attribute-contents+ `]`
-                            | `{` dialect-attribute-contents+ `}`
-                            | [^[]<>(){}\0]+
-        """
-        return self.expect(self.parse_optional_attribute, "attribute expected")
-
-    def parse_optional_attribute(self) -> Attribute | None:
-        """
-        Parse an xDSL attribute, if present.
-        An attribute is either a builtin attribute, which can have various format,
-        or a dialect attribute, with the following format:
-            dialect-attr  ::= `!` attr-name (`<` dialect-attr-contents+ `>`)?
-            attr-name     ::= bare-id
-            dialect-attr-contents ::= `<` dialect-attribute-contents+ `>`
-                            | `(` dialect-attribute-contents+ `)`
-                            | `[` dialect-attribute-contents+ `]`
-                            | `{` dialect-attribute-contents+ `}`
-                            | [^[]<>(){}\0]+
-        """
-        if (token := self._parse_optional_token(Token.Kind.HASH_IDENT)) is not None:
-            return self._parse_dialect_type_or_attribute_inner(token.text[1:], False)
-        return self._parse_optional_builtin_attr()
-
-    def _parse_dialect_type_or_attribute_inner(
-        self, attr_name: str, is_type: bool = True
-    ) -> Attribute:
-        """
-        Parse the contents of a dialect type or attribute, with format:
-            dialect-attr-contents ::= `<` dialect-attribute-contents+ `>`
-                                    | `(` dialect-attribute-contents+ `)`
-                                    | `[` dialect-attribute-contents+ `]`
-                                    | `{` dialect-attribute-contents+ `}`
-                                    | [^[]<>(){}\0]+
-        The contents will be parsed by a user-defined parser, or by a generic parser
-        if the dialect attribute/type is not registered.
-        """
-        attr_def = self.ctx.get_optional_attr(
-            attr_name,
-            create_unregistered_as_type=is_type,
-        )
-        if attr_def is None:
-            self.raise_error(f"'{attr_name}' is not registered")
-
-        # Pass the task of parsing parameters on to the attribute/type definition
-        if issubclass(attr_def, UnregisteredAttr):
-            body = self._parse_unregistered_attr_body()
-            return attr_def(attr_name, is_type, body)
-        if issubclass(attr_def, ParametrizedAttribute):
-            param_list = attr_def.parse_parameters(self)
-            return attr_def.new(param_list)
-        if issubclass(attr_def, Data):
-            self.parse_punctuation("<")
-            param: Any = attr_def.parse_parameter(self)
-            self.parse_punctuation(">")
-            return cast(Data[Any], attr_def(param))
-        assert False, "Attributes are either ParametrizedAttribute or Data."
-
     def _register_ssa_definition(
         self, name: str, values: Sequence[SSAValue], span: Span
     ) -> None:
@@ -482,6 +418,19 @@ class Parser(AttrParser):
             for val in values:
                 val.name_hint = name
 
+    @dataclass
+    class UnresolvedArgument:
+        """
+        A block argument parsed from the assembly.
+        Arguments should be parsed by `parse_argument` or `parse_optional_argument`.
+        """
+
+        name: Span
+        """The name as displayed in the assembly."""
+
+        def resolve(self, type: Attribute) -> Parser.Argument:
+            return Parser.Argument(self.name, type)
+
     @dataclass
     class Argument:
         """
@@ -492,10 +441,27 @@ class Parser(AttrParser):
         name: Span
         """The name as displayed in the assembly."""
 
-        type: Attribute | None
+        type: Attribute
         """The type of the argument, if any."""
 
-    def parse_optional_argument(self, expect_type: bool = True) -> Argument | None:
+    @overload
+    def parse_optional_argument(
+        self, expect_type: Literal[True] = True
+    ) -> Argument | None: ...
+
+    @overload
+    def parse_optional_argument(
+        self, expect_type: Literal[False]
+    ) -> UnresolvedArgument | None: ...
+
+    @overload
+    def parse_optional_argument(
+        self, expect_type: bool = True
+    ) -> UnresolvedArgument | Argument | None: ...
+
+    def parse_optional_argument(
+        self, expect_type: bool = True
+    ) -> UnresolvedArgument | Argument | None:
         """
         Parse a block argument, if present, with format:
           arg ::= percent-id `:` type
@@ -503,18 +469,32 @@ class Parser(AttrParser):
         """
 
         # The argument name
-        name_token = self._parse_optional_token(Token.Kind.PERCENT_IDENT)
+        name_token = self._parse_optional_token(MLIRTokenKind.PERCENT_IDENT)
         if name_token is None:
             return None
 
         # The argument type
-        type = None
         if expect_type:
             self.parse_punctuation(":", " after block argument name!")
             type = self.parse_type()
-        return self.Argument(name_token.span, type)
+            return self.Argument(name_token.span, type)
+        else:
+            return self.UnresolvedArgument(name_token.span)
 
-    def parse_argument(self, expect_type: bool = True) -> Argument:
+    @overload
+    def parse_argument(self, *, expect_type: Literal[True] = True) -> Argument: ...
+
+    @overload
+    def parse_argument(self, *, expect_type: Literal[False]) -> UnresolvedArgument: ...
+
+    @overload
+    def parse_argument(
+        self, *, expect_type: bool = True
+    ) -> UnresolvedArgument | Argument: ...
+
+    def parse_argument(
+        self, *, expect_type: bool = True
+    ) -> UnresolvedArgument | Argument:
         """
         Parse a block argument with format:
           arg ::= percent-id `:` type
@@ -556,14 +536,12 @@ class Parser(AttrParser):
         # Since the entry block cannot be jumped to, this is fine.
         if arguments is not None:
             # Check that the provided arguments have types.
-            if any(arg.type is None for arg in arguments):
-                raise ValueError("provided entry block arguments must have a type")
-            arg_types = cast(list[Attribute], [arg.type for arg in arguments])
+            arg_types = [arg.type for arg in arguments]
 
             # Check that the entry block has no label.
             # Since a multi-block region block must have a terminator, there isn't a
             # possibility of having an empty entry block, and thus parsing the label directly.
-            if self._current_token.kind == Token.Kind.CARET_IDENT:
+            if self._current_token.kind == MLIRTokenKind.CARET_IDENT:
                 self.raise_error("invalid block name in region with named arguments")
 
             # Set the block arguments in the context
@@ -577,8 +555,8 @@ class Parser(AttrParser):
 
         # If no arguments was provided, parse the entry block if present.
         elif self._current_token.kind not in (
-            Token.Kind.CARET_IDENT,
-            Token.Kind.R_BRACE,
+            MLIRTokenKind.CARET_IDENT,
+            MLIRTokenKind.R_BRACE,
         ):
             block = Block()
             self._parse_block_body(block)
@@ -595,8 +573,8 @@ class Parser(AttrParser):
             region.add_block(block)
 
         # Finally, check that all forward block references have been resolved.
-        if len(self.forward_block_references) > 0:
-            pos = self.lexer.pos
+        if self.forward_block_references:
+            pos = self.pos
             raise MultipleSpansParseError(
                 Span(pos, pos + 1, self.lexer.input),
                 "region ends with missing block declarations for block(s) {}".format(
@@ -606,9 +584,7 @@ class Parser(AttrParser):
                 [
                     (
                         span,
-                        'reference to block "{}" without implementation'.format(
-                            span.text
-                        ),
+                        f'reference to block "{span.text}" without implementation',
                     )
                     for span in itertools.chain(*self.forward_block_references.values())
                 ],
@@ -641,7 +617,7 @@ class Parser(AttrParser):
         Parse a list of regions with format:
            regions-list ::= `(` region (`,` region)* `)`
         """
-        if self._current_token.kind == Token.Kind.L_PAREN:
+        if self._current_token.kind == MLIRTokenKind.L_PAREN:
             return self.parse_comma_separated_list(
                 self.Delimiter.PAREN, self.parse_region, " in operation region list"
             )
@@ -661,10 +637,25 @@ class Parser(AttrParser):
         dictionary, and usually correspond to the names of the attributes that are
         already passed through the operation custom assembly format.
         """
-        begin_pos = self.lexer.pos
         if self.parse_optional_keyword("attributes") is None:
             return None
+        return self.parse_optional_attr_dict_with_reserved_attr_names(
+            reserved_attr_names
+        )
+
+    def parse_optional_attr_dict_with_reserved_attr_names(
+        self, reserved_attr_names: Iterable[str] = ()
+    ) -> DictionaryAttr | None:
+        """
+        Parse a dictionary attribute if present.
+        This is intended to be used in operation custom assembly format.
+        `reserved_attr_names` contains names that should not be present in the attribute
+        dictionary, and usually correspond to the names of the attributes that are
+        already passed through the operation custom assembly format.
+        """
+        begin_pos = self.lexer.pos
         attr = self._parse_builtin_dict_attr()
+
         for reserved_name in reserved_attr_names:
             if reserved_name in attr.data:
                 self.raise_error(
@@ -676,10 +667,11 @@ class Parser(AttrParser):
 
     def parse_optional_operation(self) -> Operation | None:
         """
-        Parse an operation, if present, with format:
+        Parse an operation with format:
             operation             ::= op-result-list? (generic-operation | custom-operation)
             generic-operation     ::= string-literal `(` value-use-list? `)`  successor-list?
-                                      region-list? dictionary-attribute? `:` function-type
+                                      properties? region-list? dictionary-attribute? `:`
+                                      function-type location?
             custom-operation      ::= bare-id custom-operation-format
             op-result-list        ::= op-result (`,` op-result)* `=`
             op-result             ::= value-id (`:` integer-literal)
@@ -687,11 +679,12 @@ class Parser(AttrParser):
             successor             ::= caret-id (`:` block-arg-list)?
             region-list           ::= `(` region (`,` region)* `)`
             dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
+            properties            ::= `<` dictionary-attribute `>`
         """
         if self._current_token.kind not in (
-            Token.Kind.PERCENT_IDENT,
-            Token.Kind.BARE_IDENT,
-            Token.Kind.STRING_LIT,
+            MLIRTokenKind.PERCENT_IDENT,
+            MLIRTokenKind.BARE_IDENT,
+            MLIRTokenKind.STRING_LIT,
         ):
             return None
         return self.parse_operation()
@@ -701,8 +694,8 @@ class Parser(AttrParser):
         Parse an operation with format:
             operation             ::= op-result-list? (generic-operation | custom-operation)
             generic-operation     ::= string-literal `(` value-use-list? `)`  successor-list?
-                                      region-list? dictionary-attribute? `:` function-type
-                                      location?
+                                      properties? region-list? dictionary-attribute? `:`
+                                      function-type location?
             custom-operation      ::= bare-id custom-operation-format
             op-result-list        ::= op-result (`,` op-result)* `=`
             op-result             ::= value-id (`:` integer-literal)
@@ -710,22 +703,31 @@ class Parser(AttrParser):
             successor             ::= caret-id (`:` block-arg-list)?
             region-list           ::= `(` region (`,` region)* `)`
             dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
+            properties            ::= `<` dictionary-attribute `>`
         """
         # Parse the operation results
         op_loc = self._current_token.span
         bound_results = self._parse_op_result_list()
 
-        if (op_name := self._parse_optional_token(Token.Kind.BARE_IDENT)) is not None:
+        if (
+            op_name := self._parse_optional_token(MLIRTokenKind.BARE_IDENT)
+        ) is not None:
             # Custom operation format
             op_type = self._get_op_by_name(op_name.text)
+            dialect_name = op_type.dialect_name()
+            self._parser_state.dialect_stack.append(dialect_name)
             op = op_type.parse(self)
+            self._parser_state.dialect_stack.pop()
         else:
             # Generic operation format
             op_name = self.expect(
                 self.parse_optional_str_literal, "operation name expected"
             )
             op_type = self._get_op_by_name(op_name)
+            dialect_name = op_type.dialect_name()
+            self._parser_state.dialect_stack.append(dialect_name)
             op = self._parse_generic_operation(op_type)
+            self._parser_state.dialect_stack.pop()
 
         n_bound_results = sum(r[1] for r in bound_results)
         if (n_bound_results != 0) and (len(op.results) != n_bound_results):
@@ -752,11 +754,11 @@ class Parser(AttrParser):
         Raises an error if the operation is not registered, and if unregistered
         dialects are not allowed.
         """
-        op_type = self.ctx.get_optional_op(name)
-        if op_type is not None:
+        if op_type := self.ctx.get_optional_op(
+            name, dialect_stack=self._parser_state.dialect_stack
+        ):
             return op_type
-
-        self.raise_error(f"unregistered operation {name}!")
+        self.raise_error(f"Operation {name} is not registered")
 
     def _parse_op_result(self) -> tuple[Span, int]:
         """
@@ -765,15 +767,15 @@ class Parser(AttrParser):
         value tuple (by default 1).
         """
         value_token = self._parse_token(
-            Token.Kind.PERCENT_IDENT, "Expected result SSA value!"
+            MLIRTokenKind.PERCENT_IDENT, "Expected result SSA value!"
         )
-        if self._parse_optional_token(Token.Kind.COLON) is None:
+        if self._parse_optional_token(MLIRTokenKind.COLON) is None:
             return (value_token.span, 1)
 
         size_token = self._parse_token(
-            Token.Kind.INTEGER_LIT, "Expected SSA value tuple size"
+            MLIRTokenKind.INTEGER_LIT, "Expected SSA value tuple size"
         )
-        size = size_token.get_int_value()
+        size = size_token.kind.get_int_value(size_token.span)
         return (value_token.span, size)
 
     def _parse_op_result_list(self) -> list[tuple[Span, int]]:
@@ -783,7 +785,7 @@ class Parser(AttrParser):
         Each result is a tuple of the span of the SSA value name (including the `%`),
         and the size of the value tuple (by default 1).
         """
-        if self._current_token.kind == Token.Kind.PERCENT_IDENT:
+        if self._current_token.kind == MLIRTokenKind.PERCENT_IDENT:
             res = self.parse_comma_separated_list(
                 self.Delimiter.NONE, self._parse_op_result, " in operation result list"
             )
@@ -794,6 +796,25 @@ class Parser(AttrParser):
     def parse_optional_attr_dict(self) -> dict[str, Attribute]:
         return self.parse_optional_dictionary_attr_dict()
 
+    def parse_optional_properties_dict(self) -> dict[str, Attribute]:
+        """
+        Parse a property dictionary, if present, with format:
+            dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
+            properties            ::= `<` dictionary-attribute `>`
+        """
+        if self.parse_optional_punctuation("<") is None:
+            return dict()
+
+        entries = self.parse_comma_separated_list(
+            self.Delimiter.BRACES, self._parse_attribute_entry
+        )
+        self.parse_punctuation(">")
+
+        if (key := self._find_duplicated_key(entries)) is not None:
+            self.raise_error(f"Duplicate key '{key}' in properties dictionary")
+
+        return dict(entries)
+
     def resolve_operands(
         self,
         args: Sequence[UnresolvedOperand],
@@ -825,14 +846,18 @@ class Parser(AttrParser):
 
     def _parse_generic_operation(self, op_type: type[Operation]) -> Operation:
         """
-        Parse an operation with format:
+        Parse a generic operation with format:
             generic-operation     ::= string-literal `(` value-use-list? `)`  successor-list?
-                                      region-list? dictionary-attribute? `:` function-type
-                                      location?
+                                      properties? region-list? dictionary-attribute? `:`
+                                      function-type location?
+            custom-operation      ::= bare-id custom-operation-format
+            op-result-list        ::= op-result (`,` op-result)* `=`
+            op-result             ::= value-id (`:` integer-literal)
             successor-list        ::= `[` successor (`,` successor)* `]`
-            successor             ::= caret-id
+            successor             ::= caret-id (`:` block-arg-list)?
             region-list           ::= `(` region (`,` region)* `)`
             dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
+            properties            ::= `<` dictionary-attribute `>`
         """
         # Parse arguments
         args = self.parse_op_args_list()
@@ -842,11 +867,14 @@ class Parser(AttrParser):
         if successors is None:
             successors = []
 
+        # Parse attribute dictionary
+        properties = self.parse_optional_properties_dict()
+
         # Parse regions
         regions = self.parse_region_list()
 
         # Parse attribute dictionary
-        attrs = self.parse_optional_attr_dict()
+        attributes = self.parse_optional_attr_dict()
 
         self.parse_punctuation(":", "function type signature expected")
 
@@ -855,14 +883,23 @@ class Parser(AttrParser):
         # Parse function type
         func_type = self.parse_function_type()
 
-        self._parse_optional_location()
+        self.parse_optional_location()
 
         operands = self.resolve_operands(args, func_type.inputs.data, func_type_pos)
 
+        # Properties retrocompatibility :
+        # We extract properties from the attribute dictionary by name.
+        if issubclass(op_type, IRDLOperation):
+            op_def = op_type.get_irdl_definition()
+            for property_name in op_def.properties.keys():
+                if property_name in attributes and property_name not in properties:
+                    properties[property_name] = attributes.pop(property_name)
+
         return op_type.create(
             operands=operands,
             result_types=func_type.outputs.data,
-            attributes=attrs,
+            properties=properties,
+            attributes=attributes,
             successors=successors,
             regions=regions,
         )
@@ -872,13 +909,16 @@ class Parser(AttrParser):
         Parse a successor with format:
             successor      ::= caret-id
         """
-        block_token = self._parse_optional_token(Token.Kind.CARET_IDENT)
+        block_token = self._parse_optional_token(MLIRTokenKind.CARET_IDENT)
         if block_token is None:
             return None
         name = block_token.text[1:]
         if name not in self.blocks:
             self.forward_block_references[name].append(block_token.span)
-            self.blocks[name] = (Block(), None)
+            block = Block()
+            if not Block.is_default_block_name(name):
+                block.name_hint = name  # setter verifies validity
+            self.blocks[name] = (block, None)
         return self.blocks[name][0]
 
     def parse_successor(self) -> Block:
@@ -894,7 +934,7 @@ class Parser(AttrParser):
             successor-list ::= `[` successor (`,` successor)* `]`
             successor      ::= caret-id
         """
-        if self._current_token.kind != Token.Kind.L_SQUARE:
+        if self._current_token.kind != MLIRTokenKind.L_SQUARE:
             return None
         return self.parse_successors()
 
@@ -920,3 +960,69 @@ class Parser(AttrParser):
             self.parse_unresolved_operand,
             " in operation argument list",
         )
+
+    def _parse_resource(
+        self, dialect_name: str, interface: OpAsmDialectInterface
+    ) -> None:
+        key = self._parse_dialect_resource_handle(dialect_name, interface)
+        self._parse_token(MLIRTokenKind.COLON, "expected `:`")
+        value = self.parse_str_literal()
+
+        try:
+            interface.parse_resource(key, value)
+        except Exception as e:
+            self.raise_error(f"got an error when parsing a resource: {e}")
+
+    def _parse_single_dialect_resources(self) -> None:
+        dialect_name = self._parse_token(
+            MLIRTokenKind.BARE_IDENT, "Expected a dialect name"
+        )
+        self._parse_token(MLIRTokenKind.COLON, "expected `:`")
+
+        dialect = self.ctx.get_optional_dialect(dialect_name.text)
+        if dialect is None:
+            self.raise_error(f"dialect {dialect_name.text} is not registered")
+
+        interface = dialect.get_interface(OpAsmDialectInterface)
+        if not interface:
+            self.raise_error(
+                f"dialect {dialect.name} doesn't have an OpAsmDialectInterface interface"
+            )
+
+        self.parse_comma_separated_list(
+            self.Delimiter.BRACES, lambda: self._parse_resource(dialect.name, interface)
+        )
+
+    def _parse_dialect_resources(self) -> None:
+        self.parse_comma_separated_list(
+            self.Delimiter.BRACES, self._parse_single_dialect_resources
+        )
+
+    def _parse_external_resources(self) -> None:
+        raise NotImplementedError("Currently only dialect resources are supported")
+
+    def _parse_metadata_element(self) -> None:
+        resource_type = self._parse_token(
+            MLIRTokenKind.BARE_IDENT, "Expected a resource type key"
+        )
+
+        self._parse_token(MLIRTokenKind.COLON, "expected `:`")
+
+        match resource_type.text:
+            case "dialect_resources":
+                self._parse_dialect_resources()
+            case "external_resources":
+                self._parse_external_resources()
+            case _:
+                self.raise_error(
+                    f"got an unexpected key in file metadata: {resource_type.text}"
+                )
+
+    def _parse_file_metadata_dictionary(self) -> None:
+        """
+        Parse metadata section {-# ... #-} of the file.
+        Returns None since results are stored in the context object.
+        """
+        self.parse_comma_separated_list(
+            self.Delimiter.METADATA_TOKEN, self._parse_metadata_element
+        )
