@@ -1,8 +1,9 @@
 from __future__ import annotations
 
 from abc import ABC
+from collections.abc import Iterable, Sequence
 from dataclasses import dataclass
-from typing import Any, Sequence, TypeGuard, cast
+from typing import Any, TypeGuard, cast
 
 from immutabledict import immutabledict
 
@@ -55,9 +56,9 @@ class IOpResult(ISSAValue):
     def __hash__(self) -> int:
         return hash(id(self.op)) + hash(self.index)
 
-    def __eq__(self, __o: object) -> bool:
-        if isinstance(__o, IOpResult):
-            return self.op == __o.op and self.index == __o.index
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, IOpResult):
+            return self.op == other.op and self.index == other.index
         return False
 
 
@@ -71,8 +72,8 @@ class IBlockArg(ISSAValue):
     def __hash__(self) -> int:
         return hash(id(self.block)) + hash(self.index)
 
-    def __eq__(self, __o: IBlockArg) -> bool:
-        return self is __o
+    def __eq__(self, other: object) -> bool:
+        return self is other
 
     def __repr__(self) -> str:
         return "BlockArg(type:" + self.type.name + ("attached") + ")"
@@ -88,8 +89,8 @@ class IRegion:
     def __hash__(self) -> int:
         return hash(id(self))
 
-    def __eq__(self, __o: object) -> bool:
-        return self is __o
+    def __eq__(self, value: object) -> bool:
+        return self is value
 
     @property
     def block(self) -> IBlock:
@@ -126,7 +127,7 @@ class IRegion:
     @classmethod
     def from_mutable(
         cls,
-        blocks: Sequence[Block],
+        blocks: Iterable[Block],
         value_map: dict[SSAValue, ISSAValue] | None = None,
         block_map: dict[Block, IBlock] | None = None,
     ) -> IRegion:
@@ -140,12 +141,6 @@ class IRegion:
         if block_map is None:
             block_map = {}
 
-        if blocks[0].parent is None:
-            raise InvalidIRException(
-                "Cannot create an IRegion from a mutable Block "
-                "that is not attached to a Region."
-            )
-
         # adding dummy block mappings so that ops have a successor to reference
         # when the actual block is created all successor references will be moved
         # to the correct block
@@ -160,6 +155,11 @@ class IRegion:
 
         # clean up successor references to blocks for ops inside this region
         for block, imm_block in zip(blocks, region.blocks):
+            if block.parent is None:
+                raise InvalidIRException(
+                    "Cannot create an IRegion from a mutable Block "
+                    "that is not attached to a Region."
+                )
             dummy_block = block_map[block]
             for block in region.blocks:
                 for op in block.ops:
@@ -219,8 +219,8 @@ class IBlock:
     def __hash__(self) -> int:
         return id(self)
 
-    def __eq__(self, __o: object) -> bool:
-        return self is __o
+    def __eq__(self, value: object) -> bool:
+        return self is value
 
     def __repr__(self) -> str:
         return (
@@ -243,12 +243,13 @@ class IBlock:
         def is_type_seq(list: Sequence[Any]) -> TypeGuard[Sequence[Attribute]]:
             return all([isinstance(elem, Attribute) for elem in list])
 
+        block_args: Sequence[IBlockArg]
         if is_type_seq(args):
-            block_args: Sequence[IBlockArg] = [
+            block_args = [
                 IBlockArg(type, IList(()), self, idx) for idx, type in enumerate(args)
             ]
         elif is_iblock_arg_seq(args):
-            block_args: Sequence[IBlockArg] = args
+            block_args = args
             for block_arg in block_args:
                 object.__setattr__(block_arg, "block", self)
         else:
@@ -343,6 +344,7 @@ class IOperation:
     __match_args__ = ("op_type", "operands", "results", "successors", "regions")
     name: str
     op_type: type[Operation]
+    properties: immutabledict[str, Attribute]
     attributes: immutabledict[str, Attribute]
     operands: IList[ISSAValue]
     results: IList[IOpResult]
@@ -354,6 +356,7 @@ class IOperation:
         name: str,
         op_type: type[Operation],
         attributes: immutabledict[str, Attribute],
+        properties: immutabledict[str, Attribute],
         operands: Sequence[ISSAValue],
         result_types: Sequence[Attribute],
         successors: Sequence[IBlock],
@@ -362,6 +365,7 @@ class IOperation:
         object.__setattr__(self, "name", name)
         object.__setattr__(self, "op_type", op_type)
         object.__setattr__(self, "attributes", attributes)
+        object.__setattr__(self, "properties", properties)
         object.__setattr__(self, "operands", IList(operands))
         for operand in operands:
             operand._add_user(self)  # pyright: ignore[reportPrivateUsage]
@@ -391,18 +395,26 @@ class IOperation:
         operands: Sequence[ISSAValue],
         result_types: Sequence[Attribute],
         attributes: immutabledict[str, Attribute],
+        properties: immutabledict[str, Attribute],
         successors: Sequence[IBlock],
         regions: Sequence[IRegion],
     ) -> IOperation:
         return cls(
-            name, op_type, attributes, operands, result_types, successors, regions
+            name,
+            op_type,
+            properties,
+            attributes,
+            operands,
+            result_types,
+            successors,
+            regions,
         )
 
     def __hash__(self) -> int:
         return hash(id(self))
 
-    def __eq__(self, __o: object) -> bool:
-        return self is __o
+    def __eq__(self, value: object, /) -> bool:
+        return self is value
 
     @property
     def result(self) -> IOpResult:
@@ -481,6 +493,7 @@ class IOperation:
         new_op: Operation = self.op_type.create(
             operands=mutable_operands,
             result_types=[result.type for result in self.results],
+            properties=dict(self.properties),
             attributes=dict(self.attributes),
             successors=mutable_successors,
             regions=mutable_regions,
@@ -536,6 +549,7 @@ class IOperation:
         else:
             operands.extend(existing_operands)
 
+        properties: immutabledict[str, Attribute] = immutabledict(op.properties)
         attributes: immutabledict[str, Attribute] = immutabledict(op.attributes)
 
         successors: list[IBlock] = []
@@ -556,7 +570,8 @@ class IOperation:
             op.name,
             op_type,
             operands,
-            [result.type for result in op.results],
+            op.result_types,
+            properties,
             attributes,
             successors,
             regions,
