@@ -3,11 +3,11 @@
 import argparse
 import sys
 
+from xdsl.context import Context
+from xdsl.dialects import get_all_dialects
 from xdsl.dialects.irdl import DialectOp
 from xdsl.dialects.irdl.irdl_to_pyrdl import convert_dialect
-from xdsl.ir import MLContext
 from xdsl.parser import Parser
-from xdsl.tools.command_line_tool import get_all_dialects
 
 
 def main():
@@ -21,10 +21,9 @@ def main():
     arg_parser.add_argument("input_file", type=str, help="path to input file")
     args = arg_parser.parse_args()
 
-    ctx = MLContext()
-    dialects = get_all_dialects()
-    for dialect in dialects:
-        ctx.register_dialect(dialect)
+    ctx = Context()
+    for dialect_factory in get_all_dialects().values():
+        ctx.load_dialect(dialect_factory())
 
     # Parse the input file
     f = open(args.input_file)
