@@ -1,6 +1,6 @@
-from xdsl.dialects import builtin
+from xdsl.dialects import builtin, stencil
 from xdsl.dialects.experimental import dmp
-from xdsl.passes import MLContext, ModulePass
+from xdsl.passes import Context, ModulePass
 from xdsl.pattern_rewriter import (
     PatternRewriter,
     PatternRewriteWalker,
@@ -11,15 +11,18 @@ from xdsl.pattern_rewriter import (
 
 class CanonicalizeDmpSwap(RewritePattern):
     @op_type_rewrite_pattern
-    def match_and_rewrite(self, op: dmp.HaloSwapOp, rewriter: PatternRewriter, /):
-        keeps: list[dmp.HaloExchangeDecl] = []
-        if op.swaps is None:
-            return
+    def match_and_rewrite(self, op: dmp.SwapOp, rewriter: PatternRewriter, /):
+        keeps: list[dmp.ExchangeDeclarationAttr] = []
         for swap in op.swaps:
             if swap.elem_count > 0:
                 keeps.append(swap)
         if len(keeps) == 0:
-            rewriter.erase_matched_op()
+            new_result = (
+                op.input_stencil
+                if isinstance(op.input_stencil.type, stencil.TempType)
+                else None
+            )
+            rewriter.replace_matched_op([], [new_result])
         else:
             op.swaps = builtin.ArrayAttr(keeps)
 
@@ -27,5 +30,5 @@ class CanonicalizeDmpSwap(RewritePattern):
 class CanonicalizeDmpPass(ModulePass):
     name = "canonicalize-dmp"
 
-    def apply(self, ctx: MLContext, op: builtin.ModuleOp) -> None:
+    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
         PatternRewriteWalker(CanonicalizeDmpSwap()).rewrite_module(op)
