@@ -1,10 +1,14 @@
 from abc import ABC
-from dataclasses import dataclass, field
-from typing import Callable, ClassVar, Iterable, TypeVar, cast
+from collections.abc import Callable, Iterable
+from dataclasses import dataclass
+from math import prod
+from typing import ClassVar, cast
 
+from xdsl.context import Context
 from xdsl.dialects import arith, builtin, func, memref, mpi, printf, scf, stencil
+from xdsl.dialects.builtin import ContainerType
 from xdsl.dialects.experimental import dmp
-from xdsl.ir import Attribute, Block, MLContext, Operation, OpResult, Region, SSAValue
+from xdsl.ir import Attribute, Block, Operation, OpResult, Region, SSAValue
 from xdsl.passes import ModulePass
 from xdsl.pattern_rewriter import (
     GreedyRewritePatternApplier,
@@ -13,36 +17,31 @@ from xdsl.pattern_rewriter import (
     RewritePattern,
     op_type_rewrite_pattern,
 )
-from xdsl.rewriter import Rewriter
-from xdsl.transforms.experimental.dmp.decompositions import (
-    DomainDecompositionStrategy,
-    GridSlice2d,
-    HorizontalSlices2D,
-)
-from xdsl.transforms.experimental.stencil_shape_inference import (
-    StencilShapeInferencePass,
-)
-from xdsl.utils.hints import isa
-
-_T = TypeVar("_T", bound=Attribute)
+from xdsl.rewriter import InsertPoint, Rewriter
+from xdsl.transforms.experimental.convert_stencil_to_ll_mlir import StencilToMemRefType
 
 _rank_dtype = builtin.i32
 
 
 @dataclass
 class ChangeStoreOpSizes(RewritePattern):
-    strategy: DomainDecompositionStrategy
+    strategy: dmp.DomainDecompositionStrategy
 
     @op_type_rewrite_pattern
     def match_and_rewrite(self, op: stencil.StoreOp, rewriter: PatternRewriter, /):
         assert all(
-            integer_attr.data == 0 for integer_attr in op.lb.array.data
+            integer_attr.data == 0 for integer_attr in op.bounds.lb.array.data
         ), "lb must be 0"
         shape: tuple[int, ...] = tuple(
-            integer_attr.data for integer_attr in op.ub.array.data
+            integer_attr.data for integer_attr in op.bounds.ub.array.data
         )
         new_shape = self.strategy.calc_resize(shape)
-        op.ub = stencil.IndexAttr.get(*new_shape)
+        op.bounds = stencil.StencilBoundsAttr.new(
+            [
+                stencil.IndexAttr.get(*(len(new_shape) * [0])),
+                stencil.IndexAttr.get(*new_shape),
+            ]
+        )
 
 
 @dataclass
@@ -51,13 +50,18 @@ class AddHaloExchangeOps(RewritePattern):
     This rewrite adds a `stencil.halo_exchange` after each `stencil.load` op
     """
 
-    strategy: DomainDecompositionStrategy
+    strategy: dmp.DomainDecompositionStrategy
 
     @op_type_rewrite_pattern
     def match_and_rewrite(self, op: stencil.LoadOp, rewriter: PatternRewriter, /):
-        swap_op = dmp.HaloSwapOp.get(op.res)
-        swap_op.nodes = self.strategy.comm_layout()
+        swap_op = dmp.SwapOp.get(op.res, self.strategy)
+        assert swap_op.swapped_values
         rewriter.insert_op_after_matched_op(swap_op)
+        for use in tuple(op.res.uses):
+            if use.operation is swap_op:
+                continue
+            use.operation.operands[use.index] = swap_op.swapped_values
+            rewriter.handle_operation_modification(use.operation)
 
 
 @dataclass
@@ -66,20 +70,18 @@ class LowerHaloExchangeToMpi(RewritePattern):
     debug_prints: bool = False
 
     @op_type_rewrite_pattern
-    def match_and_rewrite(self, op: dmp.HaloSwapOp, rewriter: PatternRewriter, /):
-        assert op.swaps is not None
-        assert op.nodes is not None
+    def match_and_rewrite(self, op: dmp.SwapOp, rewriter: PatternRewriter, /):
         exchanges = list(op.swaps)
 
-        assert isa(op.input_stencil.type, memref.MemRefType[Attribute])
+        input_type = cast(ContainerType, op.input_stencil.type)
 
         rewriter.replace_matched_op(
             list(
                 generate_mpi_calls_for(
                     op.input_stencil,
                     exchanges,
-                    op.input_stencil.type.element_type,
-                    op.nodes,
+                    input_type.get_element_type(),
+                    op.strategy.comm_layout(),
                     emit_init=self.init,
                     emit_debug=self.debug_prints,
                 )
@@ -109,7 +111,7 @@ def _generate_single_axis_calc_and_check(
     # because my_pos + 0 is always inbounds!
     if offset_in_axis == 0:
         return (
-            [true := arith.Constant.from_int_and_width(1, 1)],
+            [true := arith.ConstantOp.from_int_and_width(1, 1)],
             pos_in_axis,
             true.result,
         )
@@ -125,14 +127,16 @@ def _generate_single_axis_calc_and_check(
 
     return (
         [
-            offset_v := arith.Constant.from_int_and_width(offset_in_axis, _rank_dtype),
-            dest := arith.Addi(pos_in_axis, offset_v),
+            offset_v := arith.ConstantOp.from_int_and_width(
+                offset_in_axis, _rank_dtype
+            ),
+            dest := arith.AddiOp(pos_in_axis, offset_v),
             # get the bound we need to check:
-            bound := arith.Constant.from_int_and_width(
+            bound := arith.ConstantOp.from_int_and_width(
                 0 if is_decrement else axis_size, _rank_dtype
             ),
             # comparison == true <=> we have a valid dest positon
-            cond_val := arith.Cmpi(dest, bound, comparison),
+            cond_val := arith.CmpiOp(dest, bound, comparison),
         ],
         dest.result,
         cond_val.result,
@@ -140,7 +144,7 @@ def _generate_single_axis_calc_and_check(
 
 
 def _grid_coords_from_rank(
-    my_rank: SSAValue, grid: dmp.NodeGrid
+    my_rank: SSAValue, grid: dmp.RankTopoAttr
 ) -> tuple[list[Operation], list[SSAValue]]:
     """
     Takes a rank and a dmp.grid, and returns operations to calculate
@@ -151,20 +155,25 @@ def _grid_coords_from_rank(
     # the nodes coordinates in grid-space
     node_pos_nd: list[SSAValue] = []
 
-    # first we translate the mpi rank into grid coordinates
-    # (reversing the row major mapping)
-    divide_by = 1
-    for size in grid.as_tuple():
-        ret_ops.extend(_div_mod(my_rank, divide_by, size))
-        divide_by *= size
-        node_pos_nd.append(ret_ops[-1].results[0])
+    shape = grid.as_tuple()
+    divisors = [prod(shape[i + 1 :]) for i in range(len(shape))]
+
+    carry = my_rank
+    for div in divisors:
+        imm = arith.ConstantOp.from_int_and_width(div, builtin.i32)
+        coord_i = arith.DivUIOp(carry, imm)
+        carry = arith.RemUIOp(carry, imm)
+
+        ret_ops.extend([imm, coord_i, carry])
+        node_pos_nd.append(coord_i.result)
+
     return ret_ops, node_pos_nd
 
 
 def _generate_dest_rank_computation(
     my_rank: SSAValue,
     offsets: tuple[int, ...],
-    grid: dmp.NodeGrid,
+    grid: dmp.RankTopoAttr,
 ) -> tuple[list[Operation], SSAValue, SSAValue]:
     """
     Takes the current rank, a tuple of offsets in grid coords, and a dmp.grid
@@ -196,81 +205,60 @@ def _generate_dest_rank_computation(
     # "and" all the condition vals
     accumulated_cond_val: SSAValue = condition_vals[0]
     for val in condition_vals[1:]:
-        cmp = arith.AndI(accumulated_cond_val, val)
+        cmp = arith.AndIOp(accumulated_cond_val, val)
         ret_ops.append(cmp)
         accumulated_cond_val = cmp.result
 
     # calculate rank of destination node from grid coords
 
-    multiply_by = grid.as_tuple()[0]
-    carry: SSAValue = dest_pos_nd[0]
+    carry: SSAValue = dest_pos_nd[-1]
 
-    # dest rank: x * 1 + y * size[x] + z * size[x] * size[y] ...
-    for pos, size in zip(dest_pos_nd[1:], grid.as_tuple()[1:]):
-        fac = arith.Constant.from_int_and_width(multiply_by, _rank_dtype)
-        val = arith.Muli(pos, fac)
-        new_carry = arith.Addi(carry, val)
-        carry = new_carry.result
-        multiply_by *= size
-        ret_ops.extend([fac, val, new_carry])
+    shape = grid.as_tuple()
+    multiples = [prod(shape[i + 1 :]) for i in range(len(shape))]
+
+    for pos, mul in zip(dest_pos_nd[:-1], multiples[:-1]):
+        val = arith.ConstantOp.from_int_and_width(mul, builtin.i32)
+        intermediate = arith.MuliOp(val, pos)
+        carry_op = arith.AddiOp(carry, intermediate)
+        carry = carry_op.result
+        ret_ops.extend([val, intermediate, carry_op])
 
     return ret_ops, carry, accumulated_cond_val
 
 
-def _div_mod(i: SSAValue, div: int, mod: int) -> list[Operation]:
-    """
-    Given (i, div, mod), generate ops that calculate (i / div) % mod
-
-    The last returned operation has the final value as it's single result.
-    """
-    # these asserts should never trigger, but I'd rather have the compiler yell at me
-    # when something goes wrong than see a spectacular runtime crash :D
-    assert div > 0, "cannot work with negatives here!"
-    assert mod > 0, "cannot work with negatives here!"
-    # make sure we operate on an integer
-    assert isinstance(i.type, builtin.IntegerType | builtin.IndexType)
-    # we can use unsigned arithmetic here, because all we do is divide by positive
-    # numbers and modulo positive numbers
-    return [
-        div_v := arith.Constant.from_int_and_width(div, i.type),
-        div_res := arith.DivUI(i, div_v),
-        mod_v := arith.Constant.from_int_and_width(mod, i.type),
-        arith.RemUI(div_res, mod_v),
-    ]
-
-
 def generate_mpi_calls_for(
     source: SSAValue,
-    exchanges: list[dmp.HaloExchangeDecl],
+    exchanges: list[dmp.ExchangeDeclarationAttr],
     dtype: Attribute,
-    grid: dmp.NodeGrid,
+    grid: dmp.RankTopoAttr,
     emit_init: bool = True,
     emit_debug: bool = False,
 ) -> Iterable[Operation]:
     # call mpi init (this will be hoisted to function level)
     if emit_init:
-        yield mpi.Init()
+        yield mpi.InitOp()
     # allocate request array
     # we need two request objects per exchange
     # one for the send, one for the recv
-    req_cnt = arith.Constant.from_int_and_width(len(exchanges) * 2, builtin.i32)
-    reqs = mpi.AllocateTypeOp.get(mpi.RequestType, req_cnt)
+    req_cnt = arith.ConstantOp.from_int_and_width(len(exchanges) * 2, builtin.i32)
+    reqs = mpi.AllocateTypeOp(mpi.RequestType, req_cnt)
     # get comm rank
-    rank = mpi.CommRank.get()
+    rank = mpi.CommRankOp()
     # define static tag of 0
-    # TODO: what is tag?
-    tag = arith.Constant.from_int_and_width(0, builtin.i32)
+    tag = arith.ConstantOp.from_int_and_width(0, builtin.i32)
 
     yield from (req_cnt, reqs, rank, tag)
 
-    recv_buffers: list[tuple[dmp.HaloExchangeDecl, memref.Alloc, SSAValue]] = []
+    recv_buffers: list[
+        tuple[dmp.ExchangeDeclarationAttr, memref.AllocOp, SSAValue]
+    ] = []
 
     for i, ex in enumerate(exchanges):
         # generate a temp buffer to store the data in
         reduced_size = [i for i in ex.size if i != 1]
-        alloc_outbound = memref.Alloc.get(dtype, 64, reduced_size)
+        alloc_outbound = memref.AllocOp.get(dtype, 64, reduced_size)
         alloc_outbound.memref.name_hint = f"send_buff_ex{i}"
-        alloc_inbound = memref.Alloc.get(dtype, 64, reduced_size)
+        alloc_inbound = memref.AllocOp.get(dtype, 64, reduced_size)
         alloc_inbound.memref.name_hint = f"recv_buff_ex{i}"
         yield from (alloc_outbound, alloc_inbound)
 
@@ -283,29 +271,29 @@ def generate_mpi_calls_for(
         recv_buffers.append((ex, alloc_inbound, is_in_bounds))
 
         # get two unique indices
-        cst_i = arith.Constant.from_int_and_width(i, builtin.i32)
-        cst_in = arith.Constant.from_int_and_width(i + len(exchanges), builtin.i32)
+        cst_i = arith.ConstantOp.from_int_and_width(i, builtin.i32)
+        cst_in = arith.ConstantOp.from_int_and_width(i + len(exchanges), builtin.i32)
         yield from (cst_i, cst_in)
         # from these indices, get request objects
-        req_send = mpi.VectorGetOp.get(reqs, cst_i)
-        req_recv = mpi.VectorGetOp.get(reqs, cst_in)
+        req_send = mpi.VectorGetOp(reqs, cst_i)
+        req_recv = mpi.VectorGetOp(reqs, cst_in)
         yield from (req_send, req_recv)
 
         def then() -> Iterable[Operation]:
             # copy source area to outbound buffer
             yield from generate_memcpy(source, ex.source_area(), alloc_outbound.memref)
             # get ptr, count, dtype
-            unwrap_out = mpi.UnwrapMemrefOp.get(alloc_outbound)
+            unwrap_out = mpi.UnwrapMemRefOp(alloc_outbound)
             unwrap_out.ptr.name_hint = f"send_buff_ex{i}_ptr"
             yield unwrap_out
 
             if emit_debug:
                 yield printf.PrintFormatOp(
-                    f"Rank {{}}: sending {ex.source_area()} -> {{}}", rank, dest_rank
+                    f"Rank {{}}: sending {ex.source_area()} -> {{}}\n", rank, dest_rank
                 )
 
             # isend call
-            yield mpi.Isend.get(
+            yield mpi.IsendOp(
                 unwrap_out.ptr,
                 unwrap_out.len,
                 unwrap_out.type,
@@ -315,11 +303,11 @@ def generate_mpi_calls_for(
             )
 
             # get ptr for receive buffer
-            unwrap_in = mpi.UnwrapMemrefOp.get(alloc_inbound)
+            unwrap_in = mpi.UnwrapMemRefOp(alloc_inbound)
             unwrap_in.ptr.name_hint = f"recv_buff_ex{i}_ptr"
             yield unwrap_in
             # Irecv call
-            yield mpi.Irecv.get(
+            yield mpi.IrecvOp(
                 unwrap_in.ptr,
                 unwrap_in.len,
                 unwrap_in.type,
@@ -327,16 +315,16 @@ def generate_mpi_calls_for(
                 tag,
                 req_recv,
             )
-            yield scf.Yield.get()
+            yield scf.YieldOp()
 
         def else_() -> Iterable[Operation]:
             # set the request object to MPI_REQUEST_NULL s.t. they are ignored
             # in the waitall call
-            yield mpi.NullRequestOp.get(req_send)
-            yield mpi.NullRequestOp.get(req_recv)
-            yield scf.Yield.get()
+            yield mpi.NullRequestOp(req_send)
+            yield mpi.NullRequestOp(req_recv)
+            yield scf.YieldOp()
 
-        yield scf.If.get(
+        yield scf.IfOp(
             is_in_bounds,
             [],
             Region([Block(then())]),
@@ -344,11 +332,11 @@ def generate_mpi_calls_for(
         )
 
     # wait for all calls to complete
-    yield mpi.Waitall.get(reqs.result, req_cnt.result)
+    yield mpi.WaitallOp(reqs.result, req_cnt.result)
 
     # start shuffling data into the main memref again
     for ex, buffer, cond_val in recv_buffers:
-        yield scf.If.get(
+        yield scf.IfOp(
             cond_val,
             [],
             Region(
@@ -364,21 +352,24 @@ def generate_mpi_calls_for(
                         )
                         + [
                             printf.PrintFormatOp(
-                                f"Rank {{}} receiving from {ex.neighbor}",
+                                f"Rank {{}} receiving from {ex.neighbor}\n",
                                 rank,
                             )
                         ]
                         * (1 if emit_debug else 0)
-                        + [scf.Yield.get()]
+                        + [scf.YieldOp()]
                     )
                 ]
             ),
-            Region([Block([scf.Yield.get()])]),
+            Region([Block([scf.YieldOp()])]),
         )
 
 
 def generate_memcpy(
-    field: SSAValue, ex: dmp.HaloExchangeDecl, buffer: SSAValue, receive: bool = False
+    field: SSAValue,
+    ex: dmp.ExchangeDeclarationAttr,
+    buffer: SSAValue,
+    receive: bool = False,
 ) -> list[Operation]:
     """
     This function generates a memcpy routine to copy over the parts
@@ -388,10 +379,23 @@ def generate_memcpy(
     `field` as specified by `ex`
 
     """
-    assert isa(field.type, memref.MemRefType[Attribute])
+    field_type = cast(stencil.FieldType[Attribute], field.type)
+    assert isinstance(field_type.bounds, stencil.StencilBoundsAttr)
+    memref_type = StencilToMemRefType(field_type)
 
-    subview = memref.Subview.from_static_parameters(
-        field, field.type, ex.offset, ex.size, [1] * len(ex.offset), reduce_rank=True
+    uc = builtin.UnrealizedConversionCastOp.get([field], result_type=[memref_type])
+
+    memref_val = uc.results[0]
+
+    offset = stencil.IndexAttr.get(*ex.offset) - field_type.bounds.lb
+
+    subview = memref.SubviewOp.from_static_parameters(
+        memref_val,
+        memref_type,
+        tuple(offset),
+        ex.size,
+        [1] * len(ex.offset),
+        reduce_rank=True,
     )
     if receive:
         copy = memref.CopyOp(buffer, subview)
@@ -399,6 +403,7 @@ def generate_memcpy(
         copy = memref.CopyOp(subview, buffer)
 
     return [
+        uc,
         subview,
         copy,
     ]
@@ -428,11 +433,13 @@ class MpiLoopInvariantCodeMotion:
 
     def rewrite(
         self,
-        op: memref.Alloc
-        | mpi.CommRank
-        | mpi.AllocateTypeOp
-        | mpi.UnwrapMemrefOp
-        | mpi.Init,
+        op: (
+            memref.AllocOp
+            | mpi.CommRankOp
+            | mpi.AllocateTypeOp
+            | mpi.UnwrapMemRefOp
+            | mpi.InitOp
+        ),
         rewriter: Rewriter,
         /,
     ):
@@ -441,11 +448,11 @@ class MpiLoopInvariantCodeMotion:
         self.seen_ops.add(op)
 
         # memref unwraps can always be moved to their allocation
-        if isinstance(op, mpi.UnwrapMemrefOp) and isinstance(
-            op.ref.owner, memref.Alloc
+        if isinstance(op, mpi.UnwrapMemRefOp) and isinstance(
+            op.ref.owner, memref.AllocOp
         ):
             op.detach()
-            rewriter.insert_op_after(op.ref.owner, op)
+            rewriter.insert_op(op, InsertPoint.after(op.ref.owner))
             return
 
         base = op
@@ -467,7 +474,7 @@ class MpiLoopInvariantCodeMotion:
             return
 
         # if we move an mpi.init, generate a finalize()!
-        if isinstance(op, mpi.Init):
+        if isinstance(op, mpi.InitOp):
             # ignore multiple inits
             if parent in self.has_init:
                 rewriter.erase_op(op)
@@ -477,21 +484,21 @@ class MpiLoopInvariantCodeMotion:
             block = parent.regions[0].blocks[-1]
             return_op = block.last_op
             assert return_op is not None
-            rewriter.insert_op_before(return_op, mpi.Finalize())
+            rewriter.insert_op(mpi.FinalizeOp(), InsertPoint.before(return_op))
 
         ops = list(collect_args_recursive(op))
         for found_op in ops:
             found_op.detach()
-            rewriter.insert_op_before(base, found_op)
+            rewriter.insert_op(found_op, InsertPoint.before(base))
 
     def get_matcher(
         self,
         worklist: list[
-            memref.Alloc
-            | mpi.CommRank
+            memref.AllocOp
+            | mpi.CommRankOp
             | mpi.AllocateTypeOp
-            | mpi.UnwrapMemrefOp
-            | mpi.Init
+            | mpi.UnwrapMemRefOp
+            | mpi.InitOp
         ],
     ) -> Callable[[Operation], None]:
         """
@@ -502,11 +509,11 @@ class MpiLoopInvariantCodeMotion:
         def match(op: Operation):
             if isinstance(
                 op,
-                memref.Alloc
-                | mpi.CommRank
+                memref.AllocOp
+                | mpi.CommRankOp
                 | mpi.AllocateTypeOp
-                | mpi.UnwrapMemrefOp
-                | mpi.Init,
+                | mpi.UnwrapMemRefOp
+                | mpi.InitOp,
             ):
                 worklist.append(op)
 
@@ -521,11 +528,11 @@ class MpiLoopInvariantCodeMotion:
         """
         # collect all ops that should be rewritten
         worklist: list[
-            memref.Alloc
-            | mpi.CommRank
+            memref.AllocOp
+            | mpi.CommRankOp
             | mpi.AllocateTypeOp
-            | mpi.UnwrapMemrefOp
-            | mpi.Init
+            | mpi.UnwrapMemRefOp
+            | mpi.InitOp
         ] = list()
         matcher = self.get_matcher(worklist)
         for o in op.walk():
@@ -537,7 +544,7 @@ class MpiLoopInvariantCodeMotion:
             self.rewrite(matched_op, rewriter)
 
 
-_LOOP_INVARIANT_OPS = (arith.Constant, arith.Addi, arith.Muli)
+_LOOP_INVARIANT_OPS = (arith.ConstantOp, arith.AddiOp, arith.MuliOp)
 
 
 def can_loop_invariant_code_move(op: Operation):
@@ -571,103 +578,48 @@ def collect_args_recursive(op: Operation) -> Iterable[Operation]:
     yield op
 
 
-@dataclass
-class DmpSwapShapeInference:
-    """
-    Not a rewrite pattern, as it's a bit more involved.
-
-    This is applied after stencil shape inference has run. It will find the
-    HaloSwapOps again, and use the results of the shape inference pass
-    to attach the swap declarations.
-    """
-
-    strategy: DomainDecompositionStrategy
-    rewriter: Rewriter = field(default_factory=Rewriter)
-
-    def match_and_rewrite(self, op: dmp.HaloSwapOp):
-        core_lb: stencil.IndexAttr | None = None
-        core_ub: stencil.IndexAttr | None = None
-
-        for use in op.input_stencil.uses:
-            if not isinstance(use.operation, stencil.ApplyOp):
-                continue
-            assert use.operation.res
-            res_type = cast(stencil.TempType[Attribute], use.operation.res[0].type)
-            assert isinstance(res_type.bounds, stencil.StencilBoundsAttr)
-            core_lb = res_type.bounds.lb
-            core_ub = res_type.bounds.ub
-            break
-
-        # this shouldn't have changed since the op was created!
-        temp = op.input_stencil.type
-        assert isa(temp, stencil.TempType[Attribute])
-        assert isinstance(temp.bounds, stencil.StencilBoundsAttr)
-        buff_lb = temp.bounds.lb
-        buff_ub = temp.bounds.ub
-
-        # fun fact: pyright does not understand this:
-        # assert None not in (core_lb, core_ub, buff_lb, buff_ub)
-        assert core_lb is not None
-        assert core_ub is not None
-        assert buff_lb is not None
-        assert buff_ub is not None
-
-        op.swaps = builtin.ArrayAttr(
-            self.strategy.halo_exchange_defs(
-                dmp.HaloShapeInformation.from_index_attrs(
-                    buff_lb=buff_lb,
-                    core_lb=core_lb,
-                    buff_ub=buff_ub,
-                    core_ub=core_ub,
-                )
-            )
-        )
-
-    def apply(self, module: builtin.ModuleOp):
-        for op in module.walk():
-            if isinstance(op, dmp.HaloSwapOp):
-                self.match_and_rewrite(op)
-
-
-@dataclass
+@dataclass(frozen=True)
 class DmpDecompositionPass(ModulePass, ABC):
     """
     Represents a pass that takes a strategy as input
     """
 
-    STRATEGIES: ClassVar[dict[str, type[DomainDecompositionStrategy]]] = {
-        "2d-horizontal": HorizontalSlices2D,
-        "2d-grid": GridSlice2d,
+
+@dataclass(frozen=True)
+class DistributeStencilPass(DmpDecompositionPass):
+    """
+    Decompose a stencil to apply to a local domain.
+
+    This pass applies stencil shape inference!
+    """
+
+    name = "distribute-stencil"
+
+    STRATEGIES: ClassVar[dict[str, type[dmp.GridSlice2dAttr | dmp.GridSlice3dAttr]]] = {
+        "2d-grid": dmp.GridSlice2dAttr,
+        "3d-grid": dmp.GridSlice3dAttr,
     }
 
-    strategy: str
-
-    slices: list[int]
+    slices: tuple[int, ...]
     """
     Number of slices to decompose the input into
     """
 
-    def get_strategy(self) -> DomainDecompositionStrategy:
-        if self.strategy not in self.STRATEGIES:
-            raise ValueError(f"Unknown strategy: {self.strategy}")
-        return self.STRATEGIES[self.strategy](self.slices)
-
-
-@dataclass
-class GlobalStencilToLocalStencil2DHorizontal(DmpDecompositionPass):
+    strategy: str
     """
-    Decompose a stencil to apply to a local domain.
-
-    This pass *replaces* stencil shape inference in a
-    pass pipeline!
+    Name of the decomposition strategy to use, see STRATEGIES property for options
     """
 
-    name = "dmp-decompose-2d"
-
     restrict_domain: bool = True
+    """
+    Apply the domain restriction (i.e. change the stencil.apply to operate on the
+    local domain. If false, it assumes that the generated code is already local)
+    """
 
-    def apply(self, ctx: MLContext, op: builtin.ModuleOp) -> None:
-        strategy = self.get_strategy()
+    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
+        if self.strategy not in self.STRATEGIES:
+            raise ValueError(f"Unknown strategy: {self.strategy}")
+        strategy = self.STRATEGIES[self.strategy](self.slices)
 
         rewrites: list[RewritePattern] = [
             AddHaloExchangeOps(strategy),
@@ -681,21 +633,16 @@ class GlobalStencilToLocalStencil2DHorizontal(DmpDecompositionPass):
             apply_recursively=False,
         ).rewrite_module(op)
 
-        # run the shape inference pass
-        StencilShapeInferencePass().apply(ctx, op)
 
-        DmpSwapShapeInference(strategy).apply(op)
-
-
-@dataclass
-class LowerHaloToMPI(ModulePass):
+@dataclass(frozen=True)
+class DmpToMpiPass(ModulePass):
     name = "dmp-to-mpi"
 
     mpi_init: bool = True
 
     generate_debug_prints: bool = False
 
-    def apply(self, ctx: MLContext, op: builtin.ModuleOp) -> None:
+    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
         PatternRewriteWalker(
             GreedyRewritePatternApplier(
                 [
