@@ -2,6 +2,7 @@
 
 This module contains all custom exceptions used by xDSL.
 """
+
 from __future__ import annotations
 
 import typing
@@ -11,7 +12,21 @@ from typing import Any
 if typing.TYPE_CHECKING:
     from xdsl.ir import Attribute
     from xdsl.parser import Span
-    from xdsl.utils.parse_pipeline import Token
+    from xdsl.utils.parse_pipeline import SpecToken
+
+
+class UnregisteredConstructException(Exception):
+    """
+    An exception raised when a dialect, operation, type,
+    or attribute is not registered.
+    """
+
+
+class AlreadyRegisteredConstructException(Exception):
+    """
+    An exception raised when a dialect, operation, type,
+    or attribute is registered twice.
+    """
 
 
 class DiagnosticException(Exception):
@@ -22,22 +37,58 @@ class VerifyException(DiagnosticException):
     pass
 
 
+class PassFailedException(DiagnosticException):
+    """
+    A diagnostic error which can be raised during the execution of a pass, used to
+    signify that the pass did not succeed.
+    """
+
+    pass
+
+
 class PyRDLError(Exception):
-    pass
+    """
+    An error in our IRDL eDSL.
+    """
 
 
-class PyRDLOpDefinitionError(Exception):
-    pass
+class PyRDLOpDefinitionError(PyRDLError):
+    """
+    An error in the Operation definition eDSL.
+    """
 
 
-class PyRDLAttrDefinitionError(Exception):
-    pass
+class PyRDLAttrDefinitionError(PyRDLError):
+    """
+    An error in the Attribute definition eDSL.
+    """
+
+
+class PyRDLTypeError(TypeError, PyRDLError):
+    """
+    A type error in our IRDL eDSL.
+    """
 
 
 class InvalidIRException(Exception):
     pass
 
 
+class ShrinkException(Exception):
+    """
+    Exception for test case reduction when used in conjunction with the [Shrink Ray](https://github.com/DRMacIver/shrinkray)
+    reducer.
+
+    To find a reduced version of a test case, raise this exception on the line of code you want to hit,
+    and pass the `--shrink` argument to `xdsl-opt`, by changing its invocation from:
+    `xdsl-opt input_file.mlir -p my,pass,pipeline`
+    to:
+    `shrinkray "xdsl-opt -p my,pass,pipeline --shrink" input_file.mlir`.
+    """
+
+    pass
+
+
 class InterpretationError(Exception):
     """
     An error that can be raised during interpretation, or Interpreter setup.
@@ -54,7 +105,7 @@ class BuilderNotFoundException(Exception):
     """
 
     attribute: type[Attribute]
-    args: tuple[Any]
+    args: tuple[Any, ...]
 
     def __str__(self) -> str:
         return (
@@ -68,10 +119,7 @@ class ParseError(Exception):
     span: Span
     msg: str
 
-    def __repr__(self) -> str:
-        return self.span.print_with_context(self.msg)
-
-    def with_context(self) -> str:
+    def __str__(self) -> str:
         return self.span.print_with_context(self.msg)
 
 
@@ -80,16 +128,17 @@ class MultipleSpansParseError(ParseError):
     ref_text: str | None
     refs: list[tuple[Span, str | None]]
 
-    def __repr__(self) -> str:
-        res = super().__repr__() + "\n"
-        res += self.ref_text or "With respect to:\n"
+    def __str__(self) -> str:
+        res = self.span.print_with_context(self.msg)
+        if self.ref_text is not None:
+            res += self.ref_text + "\n"
         for span, msg in self.refs:
             res += span.print_with_context(msg) + "\n"
         return res
 
 
 class PassPipelineParseError(BaseException):
-    def __init__(self, token: Token, msg: str):
+    def __init__(self, token: SpecToken, msg: str):
         super().__init__(
             "Error parsing pass pipeline specification:\n"
             + token.span.print_with_context(msg)
