@@ -1,26 +1,33 @@
+import types
+from collections.abc import Iterable, Sequence
 from inspect import isclass
-from types import UnionType
 from typing import (
+    TYPE_CHECKING,
     Annotated,
     Any,
     Generic,
-    Iterable,
     Literal,
     TypeGuard,
-    TypeVar,
     Union,
     cast,
     get_args,
     get_origin,
 )
 
-from xdsl.ir import ParametrizedAttribute
+from typing_extensions import TypeVar
+
+from xdsl.ir import ParametrizedAttribute, SSAValue
 from xdsl.utils.exceptions import VerifyException
 
+if TYPE_CHECKING:
+    from typing_extensions import TypeForm
+
 _T = TypeVar("_T")
 
 
-def isa(arg: Any, hint: type[_T]) -> TypeGuard[_T]:
+def isa(arg: Any, hint: "TypeForm[_T]") -> TypeGuard[_T]:
+    from xdsl.irdl import ConstraintContext
+
     """
     Check if `arg` is of the type described by `hint`.
     For now, only lists, dictionaries, unions,
@@ -71,26 +78,39 @@ def isa(arg: Any, hint: type[_T]) -> TypeGuard[_T]:
             for key, value in arg_dict.items()
         )
 
-    if origin in [Union, UnionType]:
+    if origin in [Union, types.UnionType]:
         return any(isa(arg, union_arg) for union_arg in get_args(hint))
 
     if origin is Literal:
         return arg in get_args(hint)
 
+    if (origin is not None) and issubclass(origin, Sequence):
+        if not isinstance(arg, Sequence):
+            return False
+        arg_seq = cast(Sequence[Any], arg)
+        (elem_hint,) = get_args(hint)
+        return all(isa(elem, elem_hint) for elem in arg_seq)
+
     from xdsl.irdl import GenericData, irdl_to_attr_constraint
 
     if (origin is not None) and issubclass(origin, GenericData | ParametrizedAttribute):
         constraint = irdl_to_attr_constraint(hint)
         try:
-            constraint.verify(arg, {})
+            constraint.verify(arg, ConstraintContext())
             return True
         except VerifyException:
             return False
 
+    if origin is SSAValue:
+        if not isinstance(arg, SSAValue):
+            return False
+        arg = cast(SSAValue, arg)
+        return isa(arg.type, get_args(hint)[0])
+
     raise ValueError(f"isa: unsupported type hint '{hint}' {get_origin(hint)}")
 
 
-def assert_isa(arg: Any, hint: type[_T]) -> TypeGuard[_T]:
+def assert_isa(arg: Any, hint: "TypeForm[_T]") -> TypeGuard[_T]:
     """
     Check if `arg` is of the type described by `hint`.
     For now, only lists, dictionaries, unions,
@@ -129,43 +149,77 @@ def get_type_var_from_generic_class(cls: type[Any]) -> tuple[TypeVar, ...]:
 
 def get_type_var_mapping(
     cls: type[Any],
-) -> tuple[type[Any], dict[TypeVar, Any]]:
+) -> tuple[tuple[TypeVar, ...], dict[TypeVar, Any]]:
     """
-    Given a class that specializes a generic class, return the generic class and
-    the mapping from the generic class type variables to the specialized arguments.
+    Given a Generic class, return the TypeVars used to specialize it, and the mapping
+    from the generic class type variables to the specialized arguments for all type
+    variables used in ancestor classes.
+    Raises a ValueError if the specialized arguments for the same TypeVar are not
+    consistent among superclasses.
     """
 
     if not issubclass(cls, Generic):
         raise ValueError(f"{cls} does not specialize a generic class.")
 
-    # Get the generic parent
     orig_bases: Iterable[Any] = getattr(cls, "__orig_bases__")
-    orig_bases = [
-        orig_base for orig_base in orig_bases if get_origin(orig_base) is not Generic
-    ]
-    # Do not handle more than one generic parent in the mro.
-    # It is possible to handle more than one generic parent, but
-    # the mapping of type variables will be more complex, especially for
-    # generic parents inheriting from other generic parents.
-    if len(orig_bases) != 1:
-        raise ValueError(
-            "Class cannot have more than one generic class in its mro. This "
-            "restriction may be lifted in the future.",
-            orig_bases,
-        )
+    mapping: dict[TypeVar, Any] = {}
+    args: tuple[TypeVar, ...] = ()
 
-    # Get the generic operation, and its specialized type parameters.
-    generic_parent: type[Any] = get_origin(orig_bases[0])
-    specialized_args = get_args(orig_bases[0])
+    for base in orig_bases:
+        base_origin = get_origin(base)
+        if base_origin is Generic:
+            args = get_args(base)
 
-    # Get the `TypeVar` used in the generic parent
-    generic_args = get_type_var_from_generic_class(generic_parent)
+    for base in orig_bases:
+        base_origin = get_origin(base)
+        if base_origin is Generic:
+            continue
+        base_cls = base if base_origin is None else base_origin
+        if not issubclass(base_cls, Generic):
+            continue
+        base_type_vars, base_mapping = get_type_var_mapping(cast(type[Any], base_cls))
+        base_args = get_args(base)
 
-    if len(generic_args) != len(specialized_args):
-        raise ValueError(
-            f"Generic class {generic_parent} class has {len(generic_args)} "
-            f"parameters, but {cls} specialize {len(specialized_args)} of them."
-        )
+        for k, v in zip(base_type_vars, base_args, strict=True):
+            if isinstance(v, TypeVar) and v not in args:
+                # Pyright complains if there is a Generic parent that doesn't include
+                # all the TypeVars used in later classes, so this is a valid check.
+                raise ValueError(
+                    f"Invalid definition {cls.__qualname__}, generic classes must subclass `Generic`."
+                )
+            if k is not v:
+                # Don't assign forwarded TypeVars
+                base_mapping[k] = v
 
-    type_var_mapping = dict(zip(generic_args, specialized_args))
-    return generic_parent, type_var_mapping
+        for k, v in base_mapping.items():
+            if k in mapping:
+                if v is not mapping[k]:
+                    raise ValueError(
+                        "Error extracting assignments of TypeVars of "
+                        f"{cls.__qualname__}, inconsistent assignments to {k} in "
+                        f"superclasses: {v.__qualname__}, {mapping[k].__qualname__}."
+                    )
+                continue
+            mapping[k] = v
+
+    return args, mapping
+
+
+def type_repr(obj: Any) -> str:
+    """Return the repr() of an object, special-casing types."""
+    if isinstance(obj, types.GenericAlias):
+        origin = get_origin(obj)
+        args = get_args(obj)
+        return f"{type_repr(origin)}[{', '.join(type_repr(arg) for arg in args)}]"
+    if isinstance(obj, types.UnionType):
+        args = get_args(obj)
+        return f"{'|'.join(type_repr(arg) for arg in args)}"
+    if obj is type(None):
+        return "None"
+    if obj is ...:
+        return "..."
+    if isinstance(obj, types.FunctionType):
+        return obj.__name__
+    if isinstance(obj, type):
+        return f"{obj.__qualname__}"
+    return repr(obj)
